# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Structured configuration formats such as corosync.conf"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..format import SimpleFormat
from ..protocol import Protocol
from ..utils_2to3 import basestring, bytes_enc, str_enc
from ..utils_func import apply_aggregation_preserving_passing_depth
from ..utils_prog import getenv_namespaced


class simpleconfig(SimpleFormat):
    """Structured configuration formats such as corosync.conf

    Internally ('struct'), it is structured liked this:

    SECTION  ::= tuple(TAG, OPTIONS, SECTIONS)  # [] tolerated, but fixed width
    SECTIONS ::= list-or-tuple(SECTION)
    OPTIONS  ::= list-or-tuple(OPTION)
    OPTION   ::= tuple(TAG, VALUE)  # must be tuple due to inner arragement
    TAG      ::= .*   # XXX and comment handling [^#@].*
    VALUE    ::= .*

    where SECTIONS form logical subsections (arbitrarily nested)

    Example:

    ('corosync-ONLY-INTERNAL-TAG-NOT-EXTERNALIZED-ANYWAY',
     [],
     [('totem', [('version', '2'), ('cluster_name', 'aus-cluster')], {}),
      ('nodelist',
       [],
       [('node', [('nodeid', '1'), ('ring0_addr', 'lolek.example.com')], []),
        ('node', [('nodeid', '2'), ('ring0_addr', 'bolek.example.com')], [])]),
      ('quorum',
       [('provider', 'corosync_votequorum'),
        ('expected_votes', '1'),
        ('two_node', '1')],
       [])])

    See also: https://github.com/lhh/simpleconfig
    """
    # NOTE yacc-based parser in fence-virt
    native_protocol = STRUCT = Protocol('struct')
    BYTESTRING = SimpleFormat.BYTESTRING

    # notable lexical units for input
    lbrace_i, rbrace_i, optsep_i = '{',            '}',      ':'
    # notable lexical units for output (~ pretty-printing, hence with spaces)
    lbrace_o, rbrace_o, optsep_o = ' ' + lbrace_i, rbrace_i, optsep_i + ' '
    # same for input/output
    csep = '#'

    @SimpleFormat.producing(BYTESTRING)
    def get_bytestring(self, *iodecl):
        """Externalize 'struct', that is basically, pretty print it

        For example above, the result is something like:

        totem {
            version: 2
            cluster_name: aus-cluster
        }
        nodelist {
            node {
                id: 1
                ring0_addr: lolek.example.com
            }
            node {
                id: 2
                ring0_addr: bolek.example.com
            }
        }
        quorum {
            provider: corosync_votequorum
            expected_votes: 1
            two_node: 1
        }
        """
        # try to look (indirectly) if we have a file at hand first
        ret = super(simpleconfig, self).get_bytestring(self.BYTESTRING)
        if ret is not None:
            return ret

        # fallback
        struct = self.STRUCT(protect_safe=True)
        indent, optindent = (getenv_namespaced('COROINDENT', '\t'), ) * 2
        lbrace, rbrace, optsep = self.lbrace_o, self.rbrace_o, self.optsep_o
        ret = '\n'.join(
            apply_aggregation_preserving_passing_depth(
                lambda x, d:
                    # avoid IndexError
                    x if not x or not isinstance(x[0], basestring)
                    else
                    # OPTION
                    ((d - 3) // 2 * indent + x[0] + optsep + x[1], )
                        if isinstance(x, tuple) and len(x) == 2  # dif. SECTION
                    else
                    # rest ([''] at the end is for a final newline)
                    ([(d - 3) // 2 * indent + x[0] + lbrace] if d > 1 else [])
                        + list(xxxx for xx in filter(bool, x[1:])
                               for xxx in xx for xxxx in xxx)
                        + ([(d - 3) // 2 * indent + rbrace] if d > 1 else [''])
            )(struct)
        )
        return bytes_enc(ret, 'utf-8')

    @SimpleFormat.producing(STRUCT, protect=True)
    def get_struct(self, *iodecl):
        # similar to pcs.corosync_conf._parse_section (and Corosync itself)
        lbrace, rbrace, optsep = self.lbrace_i, self.rbrace_i, self.optsep_i
        csep = self.csep
        attrs, children = [], []
        work, t = [('', attrs, children)], str_enc(self.BYTESTRING(), 'utf-8')
        while t:
            h, t = t.split('\n', 1)
            h = h.lstrip(" \t")  # otherwise the following may not work
            if h.startswith(csep):
                continue
            elif lbrace in h:
                work.append((h.split(lbrace, 1)[0].strip(), [], []))
                attrs, children = work[-1][1:]
            elif rbrace in h:
                try:
                    cur = work.pop()
                    attrs, children = work[-1][1:]
                except IndexError:
                    raise RuntimeError("Unexpected closing brace")
                children.append(cur)
            elif optsep in h:
                attrs.append(tuple(x.strip() for x in h.split(optsep, 1)))
        ret = work.pop()
        if work:
            raise RuntimeError("Missing {0} closing brace(s)".format(len(work)))
        return ret


class simpleconfig_normalized(simpleconfig):
    """Structured configuration formats such as corosync.conf, normalized form

    Here, "normalized" is a synonym to "bijectively convertible to XML".
    Trivially, any elements-attributes XML is convertible to simpleconfig,
    but because simpleconfig can carry name-/key-duplicated options, something
    not suitable for reverse options-to-attributes mapping has to be normalized
    and because this is to serve in simpleconfig2needlexml filter (hence
    affecting only? uidgid entries), we define the normalization as follows:

    1. analyze current section whether it contains repeated options
      1a. no  -> goto 2. right away
      1b. yes -> for each duplicated option but without its value duplication
                 (in that case, such issue a warning), create a new
                 "following sibling" section and move the option here on its
                 own; if the current section contains subsection,
                 issue a warning about that
    2. continue with the next section in a defined traversal

    Fair enough, we've just described the simpleconfig-normalize filter :)
    """
    pass



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from __future__ import (print_function,
                        absolute_import,  # for commands/std. library in PY2.6
                        )

"""Machinery entry point"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

import logging
from optparse import OptionParser, \
                     OptionGroup, \
                     IndentedHelpFormatter
from os.path import basename, realpath
# XXX should eventually switch/fallback to "distro" external package for
# the latter (via https://bugzilla.redhat.com/1219172#c6, but see also #c9)
from platform import system, linux_distribution
try:
    from subprocess import check_output
    getoutput = lambda cmd: check_output(cmd, shell=True)
except ImportError:
    from commands import getoutput
from sys import version

from . import version_parts, version_text, description_text
from .command_manager import CommandManager
from .completion import Completion
from .error import EC
from .facts import aliases_dist, aliases_rel, format_dists, supported_dists
from .utils import args2sgpl, head_tail, identity
from .utils_2to3 import iter_items, str_enc, xrange
from .utils_func import foreach
from .utils_prog import ExpertOption, make_options, set_logging, which

try:
    from .defaults import REPORT_BUGS
except ImportError:
    report_bugs = ()
else:
    report_bugs = ("Report bugs to", "<{0}>".format(REPORT_BUGS))
    joint = ' '.join(report_bugs) + '.'
    if len(joint) < 70:
           report_bugs = (joint, )

def _resolve_dist(sys, dist_name, *dist_ver, **kws):
    # **kws: deferred_log=None
    distro, version, recognized = dist_name, dist_ver, True
    dl = kws.get("deferred_log", None)
    dl = dl if dl is not None else []
    for fn in (lambda x: x.lower(), lambda x: aliases_dist.get(x, x), identity):
        if distro in supported_dists(sys):
            if distro != dist_name:
                dl.append((logging.INFO, "Distro `{0}' recognized as `{1}'"
                                         .format(dist_name, distro)))
            break
        distro = fn(distro)
    else:
        recognized = False
        dl.append((logging.WARNING, "Unrecognized distro `{0}' may lead to"
                                    " unexpected results".format(dist_name)))
    if not dist_ver:
        dl.append((logging.WARNING, "Missing `{0}' distro version may lead to"
                                    " unexpected results".format(dist_name)))
    elif recognized:
        version, version_rest = head_tail(version)
        aliases_ver = aliases_rel.get(distro, {})
        for fn in (lambda x: x.lower(), lambda x: aliases_ver.get(x, x), identity):
            if version != dist_ver[0]:
                dl.append((logging.INFO, "Version `{0}' recognized as `{1}'"
                                         .format(dist_ver[0], version)))
                break
            version = fn(version)
        else:
            if version.strip("0123456789."):
                dl.append((logging.WARNING, "Unrecognized non-numeric version"
                                            " `{0}' may lead to unexpected"
                                            " results".format(dist_ver[0])))
        version = args2sgpl(version, *version_rest)

    return args2sgpl(distro, *version)


_deferred_log = []
_system = system().lower()
_system_extra = (linux_distribution(full_distribution_name=0)
                    if _system == 'linux'
                 else (_system, str_enc(getoutput("/bin/freebsd-version -u"))
                                .rsplit('-', 1)[0])
                    if _system == 'freebsd'
                 else ())
if _system.endswith("bsd"):
    _system = "bsd"
_system_extra = _resolve_dist(_system, *_system_extra,
                              **dict(deferred_log=_deferred_log))


def parser_callback_help(option, opt_str, value, parser, arg=False, full=False):
    """Makes 'help' option accept 'optional option argument'"""
    if arg:
        rargs, val = parser.rargs, ''
        if rargs and not rargs[0].startswith('-'):
            val = rargs[0]
            del rargs[:1]
    else:
        val = True
    setattr(parser.values, 'help', val)
    setattr(parser.values, 'help_full', full)


def parser_callback_sys(option, opt_str, value, parser, *args, **kwargs):
    setattr(parser.values, option.dest, value.lower())


def parser_callback_dist(option, opt_str, value, parser, *args, **kwargs):
    parser.values._deferred_log = dl = getattr(parser.values,
                                               '_deferred_log', [])
    distro, version = head_tail(value.split(','))
    setattr(parser.values, option.dest,
            ','.join(_resolve_dist(parser.values.sys, distro, *version,
                                   **dict(deferred_log=dl))))


def parser_callback_list_dists(option, opt_str, value, parser):
    """Makes 'list-dists' option accept 'optional option argument'"""
    val = 0
    rargs = parser.rargs
    if rargs and rargs[0]:
        try:
            val = int(rargs[0])
            del rargs[:1]
        except ValueError:
            pass
    setattr(parser.values, 'list', 'dists_{0}'.format(val))


opts_common = (
    (('--sys', ), dict(
        type='string',  # for dest -> default
        action='callback',
        callback=parser_callback_sys,
        default=_system,
        expert=True,
        help="override autodetected system [%default]"
    )),
    (('--dist', ), dict(
        type='string',  # for dest -> default
        action='callback',
        callback=parser_callback_dist,
        default=','.join(_system_extra),
        help="override target distro (for SYS=linux; see --list-dists)"
             " [%default]"
    )),
    (('-q', '--quiet', ), dict(
        action='store_true',
        help="refrain from unnecesary messages (usually on stderr)"
    )),
    (('--color', ), dict(
        metavar="[WHEN]",
        action='store',
        dest='color',
        default='auto',
        type='choice',
        choices=('auto', 'never', 'always'),
        help="colorize messages if available [%default out of %choices]"
    )),
    (('-d', '--debug'), dict(
        action='store_const',
        dest='loglevel',
        const='DEBUG',
        help="shortcut for --loglevel=DEBUG"
    )),
    (('--logfile', ), dict(
        metavar="FILE",
        action='store',
        dest='logfile',
        default='',
        help="specify log file (instead of stderr)"
    )),
    (('--loglevel', ), dict(
        metavar="LEVEL",
        action='store',
        dest='loglevel',
        default=logging.getLevelName(logging.WARNING),
        type='choice',
        choices=list(map(logging.getLevelName,
                         xrange(logging.NOTSET, logging.CRITICAL + 1,
                                logging.DEBUG - logging.NOTSET))),
        help="specify log level [%default out of %choices]"
    )),
    # TODO other logging related stuff (if any)
)

opts_main = (
    (('-h', '--help'), dict(
        metavar="[CMD]",
        type='string',
        nargs=0,  # <- we take one if suitable
        action='callback',
        callback=lambda *args: parser_callback_help(*args, arg=True),
        help="show help message (can be command-specific) and exit"
    )),
    (('-H', '--help-full'), dict(
        metavar="[CMD]",
        type='string',
        nargs=0,  # <- we take one if suitable
        action='callback',
        callback=lambda *args: parser_callback_help(*args, arg=True, full=True),
        help="full help message (can be command-specific) and exit"
    )),
    (('-l', '--list'), dict(
        action='store_const',
        const='cmds',
        help="list commands and exit"
    )),
    (('--list-dists', ), dict(
        metavar="[VERBOSITY]",
        type='string',
        nargs=0,  # <- we take one if suitable
        action='callback',
        callback=parser_callback_list_dists,
        expert=True,
        help="list explicitly supported --dist option arguments and exit"
    )),
    (('-v', '--version'), dict(
        action='store_true',
        help="show version details and exit"
    )),
    (('-e', '--ext'), dict(
        action='append',
        dest='ext_user',
        expert=True,
        help="specify dir/s (as PATH/multiplied) scanned for plugins"
    )),
    (('-s', '--skip-ext'), dict(
        action='store_true',
        dest='skip_ext',
        help="do not use standard external plugins"
    )),
    (('--completion-bash', ), dict(
        action='store_const',
        dest='completion', const='bash',
        help="generate bash completion and exit"
    )),
)

opts_nonmain = (
    (('-h', '--help'), dict(
        action='callback',
        callback=parser_callback_help,
        help="show help message and exit"
    )),
    (('-H', '--help-full'), dict(
        action='callback',
        callback=lambda *args: parser_callback_help(*args, full=True),
        help="full help message and exit"
    )),
)


class SharedHelpFormatter(IndentedHelpFormatter):
    """IndentedHelpFormatter to expand choices along defaults"""
    choices_tag = "%choices"

    def __init__(self, *args, **kwargs):
        from textwrap import fill, wrap
        IndentedHelpFormatter.__init__(self, *args, **kwargs)
        tw = type('textwrap_mock', (object, ), dict(
            fill=staticmethod(fill),
            wrap=staticmethod(lambda *args, **kwargs:
                wrap(*args, **dict(kwargs, break_on_hyphens=False)),
            )))
        self.format_option.__globals__['textwrap'] = tw

    def expand_default(self, option):
        ret = IndentedHelpFormatter.expand_default(self, option)
        if isinstance(option, ExpertOption):
            ret = "(Advanced) " + ret
        return ret.replace(self.choices_tag, ', '.join(option.choices or []))


class SharedHelpFormatterNonExpert(SharedHelpFormatter):
    """SharedHelpFormatter to filter out expert options"""
    def format_option(self, option):
        if not isinstance(option, ExpertOption):
            ret = SharedHelpFormatter.format_option(self, option)
        else:
            ret = ''
        return ret


class SharedOptionParser(OptionParser):
    """OptionParser with a dynamic on-demand help screen customization."""

    formatter_nonexpert = SharedHelpFormatterNonExpert()
    formatter_expert = SharedHelpFormatter()

    # overridden methods

    def __init__(self, **kwargs):
        # a bit awkward, but in place as a sort of memoizing
        if 'formatter' not in kwargs:
            kwargs['formatter'] = self.formatter_nonexpert
        OptionParser.__init__(self, **kwargs)
        self.formatter_expert.set_parser(self)  # ensure %default expansion
        self.description_raw = ''

    def format_description(self, formatter):
        # cf. http://bugs.python.org/issue4318
        return '\n'.join(formatter.format_description(l)
                         for l in self.get_description().split('\n\n')) \
               + (self.description_raw and '\n' + self.description_raw + '\n')

    def format_epilog(self, formatter):
        ret = '\n' + '\n'.join(formatter.format_epilog(l).strip('\n')
                               if '<http' not in l else l
                               for l in self.epilog.split('\n'))
        return ret

    # custom methods

    def format_customized_help(self, **kwargs):
        for k in ('usage', 'description', 'description_raw', 'epilog'):
            v = kwargs.pop(k, None)
            if v:
                setattr(self, k, v)
        help_full = getattr(self.values, 'help_full', None)
        if help_full in (True, False):
            self.help_full(help_full)
        return self.format_help(**kwargs)

    def add_option_group_by_args(self, *args, **kwargs):
        option_list = kwargs.pop('option_list', None)
        group = OptionGroup(self, *args, **kwargs)
        if option_list:
            group.add_options(option_list)
        self.add_option_group(group)

    def help_full(self, expert):
        assert self.formatter in (self.formatter_nonexpert,
                                  self.formatter_expert), "explicit formatter"
        self.formatter = (self.formatter_expert if expert else
                          self.formatter_nonexpert)


def run(argv=None, *args):
    """Entry point"""
    # re option parser: only one instance is used, modified along
    ec = EC.EXIT_SUCCESS
    argv = list(argv) + list(args) if argv else list(args)
    prog, argv = (argv[0], argv[1:]) if argv else ('<script>', [])
    prog_simple = basename(prog)
    prog_full = prog if prog != prog_simple else which(prog_simple, '.', '')
    prog_real = basename(realpath(prog_full))

    parser = SharedOptionParser(prog=prog_simple, add_help_option=False)
    parser.disable_interspersed_args()  # enforce ordering as per "usage"
    parser.add_option_group_by_args(
        "Standalone global options", "These take precedence over any command.",
        option_list=make_options(opts_main)
    )
    parser.add_option_group_by_args(
        "Common options",
        "Either in global (before <cmd>) or command scope (after <cmd>).",
        option_list=make_options(opts_common)
    )

    opts, args = parser.parse_args(argv)
    if prog_simple == prog_real and opts.help is None:
        # options that return/exit + don't need plugin resolutions (not help)
        if opts.version:
            loglevel = logging.getLevelName(opts.loglevel)
            msg = version_parts
            if loglevel <= logging.INFO:
                msg += ('', "Python runtime:", version)
            print(version_text(*msg))
            return ec

    logging.basicConfig()
    try:
        # only 2.7+ (despite not documented this way)
        logging.captureWarnings(True)
    except AttributeError:
        pass
    set_logging(opts)
    log = logging.getLogger(__name__)
    foreach(lambda args: log.log(*args),
            getattr(opts, '_deferred_log', _deferred_log))

    cm = CommandManager.init_lookup(ext_plugins=not opts.skip_ext,
                                    ext_plugins_user=opts.ext_user,
                                    system=opts.sys, system_extra=opts.dist)
    if prog_simple == prog_real and not opts.help \
       and (opts.list or opts.completion or not args):
        cmds = cm.pretty_cmds(ind=' ' * parser.formatter.indent_increment,
                              linesep_width=2,
                              cmds_intro="Commands"
                                         " (as available, but stable)",
                              aliases_intro="Aliases thereof"
                                            " (environment-specific):",
                              ellip_yes='; built-in only' + (
                                         ' required:' if opts.skip_ext else
                                        ', use --list to get all:'),
                              ellip=opts.skip_ext or not opts.list)
        if opts.list == 'cmds':
            print(cmds)
        elif opts.list and opts.list.startswith('dists_'):
            verbosity, acc = int(opts.list.split('dists_', 1)[1]), []
            if verbosity == 0:
                acc.extend((
                    "# boundary (or so) versions of what is explicitly"
                    " supported;",
                    "# increase verbosity to display more",
                ))
            elif verbosity >= 1:
                acc.extend((
                    "# spanning, still possibly sparse subset of what is"
                    " explicitly supported",
                    "# (all versions any change/update is tracked at)",
                ))
            print('\n'.join(acc + [format_dists(verbosity, sys=_system)]))
        elif opts.completion:
            c = Completion.get_completion(opts.completion, prog,
                                          opts_common, opts_main, opts_nonmain)
            print(c(iter_items(cm.plugins)))
        else:
            print(parser.format_customized_help(
                usage="%prog [<global option> ...] [<cmd> [<cmd option ...>]]",
                description=description_text(width=0),
                description_raw=cmds,
                epilog='\n'.join(args2sgpl(
                    "To get help for given command, just precede or follow"
                    " it with --help.",
                    *report_bugs
                ))
            ))
        return ec
    elif prog_simple != prog_real:
        args = [prog_simple] + argv

    # prepare option parser to be reused by sub-commands
    parser.enable_interspersed_args()
    modify_group = parser.get_option_group(opts_main[0][0][0])
    foreach(parser.remove_option, map(lambda x: x[0][0], opts_main))
    modify_group.set_title("Command options")
    modify_group.set_description(None)
    parser.add_options(make_options(opts_nonmain))
    parser.epilog = '\n'.join(args2sgpl(
        "Arguments to value-based `command options' can go without labels"
        " when the order wrt. parsing logic respected;"
        " skipping those backed by default values otherwise requiring"
        " specification then allowed by syntactic sugar: all can be passed"
        " as a single, first, ::-delimited argument;"
        " magic files: `-', `@DIGIT+'. `{{formula}}' in output file spec:"
        " input-backed (e.g. hash) substitution recipe."
        "  All available commands listed as `{0} --list'."
        .format(prog_simple),
        *report_bugs
    ))
    #try:
    # note that the parser carries opts and "Common options" group
    ec = cm(parser, args)
    #except Exception as e:
    #    print("FATAL: Unhandled exception: {0}".format(e))
    #    ex = EC.EXIT_FAILURE
    return ec



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl


ccsflat2cibprelude = '''\
    <!--
        apache ~ apache
     -->
    <xsl:when test="name() = 'apache'">
''' + (
        ResourceSpec('ocf:heartbeat:apache').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">

            <!-- configfile ~ (server_root + '/' +) config_file -->
            <xsl:choose>
                <xsl:when test="starts-with(@config_file, '/')">
''' + (
                    rg2hb_xsl('configfile', 'config_file', req=True)
) + '''\
                </xsl:when>
                <xsl:otherwise>
''' + (
                    rg2hb_xsl('configfile',
                              "{concat(@server_root, '/', @config_file)}",
                              req=abs)
) + '''\
                </xsl:otherwise>
            </xsl:choose>

            <!-- options ~ httpd_options (if present; + name, server_root) -->
            <nvpair id="{concat($Prefix, '-ATTRS-options')}"
                    name="options">
                <xsl:attribute name="value">
                    <xsl:value-of select="concat(
                        ' -D', @name
                    )"/>
                    <xsl:if test="@server_root">
                        <xsl:value-of select="concat(
                            ' -d &quot;', @server_root, '&quot;'
                        )"/>
                    </xsl:if>
                    <xsl:if test="@httpd_options">
                        <xsl:value-of select="concat(
                            ' ', @httpd_options
                        )"/>
                    </xsl:if>
                </xsl:attribute>
            </nvpair>
''' + (
            # see rhbz#952132
            rg2hb_xsl('httpd')
) + '''\
        </instance_attributes>

        <!-- OPERATIONS -->
        <operations>
''' + (
            rg2hb_xsl('stop', 'shutdown_wait', op=True)
) + '''\
        </operations>
    </xsl:when>
'''

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@httpd',
                     kind='B', desc='path to httpd binary'),
    artefact_cond_ra('@server_root',
                     kind='D', desc='ServerRoot'),
    artefact_cond_ra('@config_file[../@server_root]',
                     xpath="concat(../@server_root, '/', .)",
                     kind='A', desc='configuration file'),
))



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))

###

from ....utils_cib import ResourceSpec, rg2hb_xsl
from ....utils_xml import squote

ccsflat2cibprelude = ('''\
    <!--
        tomcat ~ tomcat-6
     -->
    <xsl:when test="name() = 'tomcat-6'">
''' + (
        ResourceSpec('ocf:heartbeat:tomcat').xsl_attrs
) + '''
        <xsl:comment><xsl:value-of select="concat(' ', %(note)s, ' ')"/></xsl:comment>
        <xsl:message><xsl:value-of select="concat(%(note)s)"/></xsl:message>

        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('java_home', '/usr/lib/jvm/jre', req=abs)
            +
            rg2hb_xsl('catalina_home', '{$pcscmd_tomcat_catalina_home}',
                      req=abs)
) + '''\
        </instance_attributes>

        <!-- OPERATIONS -->
        <operations>
''' + (
            rg2hb_xsl('stop', 'shutdown_wait', op=True)
) + '''\
        </operations>
    </xsl:when>
''') % dict(
    note=', '.join((
        squote("NOTE: cannot convert tomcat-6 resource directly,"
               " prefilling defaults; "),
        squote("consider manual completion based on /etc/tomcat6/tomcat6.conf ("),
        "@config_file",
        squote(") file (or switching to LSB/systemd)")
    ))
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Filter manager"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from logging import getLogger

from .error import ClufterError
from .filter import filters
from .format_manager import FormatManager
from .plugin_registry import PluginManager
from .utils_2to3 import iter_values

log = getLogger(__name__)


class FilterManagerError(ClufterError):
    pass


class FilterManager(PluginManager):
    """Class responsible to manage filters and filtering itself"""
    _default_registry = filters

    @classmethod
    def _init_plugins(cls, filters, fmt_mgr=None, **kwargs):
        log.debug("Filters before resolving: {0}".format(filters))
        if fmt_mgr is None:
            fmts = set()
            for flt in iter_values(filters):
                # XXX composite format
                for attr in ('in_format', 'out_format'):
                    fmts.add(getattr(flt, attr))
            fmt_mgr = FormatManager.init_lookup(fmts, **kwargs)
        return cls._resolve(fmt_mgr.formats, filters)

    @staticmethod
    def _resolve(formats, filters):
        for flt_name, flt_cls in tuple(filters.items()):
            ret = flt_cls(formats)
            if ret is not None:
                filters[flt_name] = ret
            else:
                filters.pop(flt_name)
        return filters

    @property
    def filters(self):
        return self._plugins

    def __call__(self, which, in_decl, **kwargs):
        flt = self._plugins[which]
        in_obj = flt.in_format.as_instance(*in_decl)
        return flt(in_obj, **kwargs)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Filters for upgrading CIB formats"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import XMLFilter
from ..utils_xslt import xslt_params

from os.path import basename, splitext


@XMLFilter.deco('cib-1', 'cib-2')
def fmt_cib_1to2(flt_ctxt, in_obj):
    self = flt_ctxt.ctxt_wrapped
    validator_spec_min = self.out_format.etree_rng_validator_proper_specs()[0]
    if validator_spec_min:
        validator_spec_min = splitext(basename(validator_spec_min))[0]
    else:
        validator_spec_min = 'pacemaker-2.0'
    return (
        'etree',
        flt_ctxt.ctxt_proceed_xslt(
            in_obj,
            def_first=xslt_params(
                cib2_min_ver=validator_spec_min
            )
        )
    )



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccsflat2cibprelude filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..facts import cluster_pcs_flatiron, component_or_state, package, system
from ..filter import XMLFilter
from ..utils_xslt import xslt_params

from os.path import join


# XXX temporary hack with plain ccs;
# check that it is indeed ccs-flat, by exploring flt_ctxt?
@XMLFilter.deco('ccs-flat', 'cib-prelude')
def ccsflat2cibprelude(flt_ctxt, in_obj):
    sys_pair = flt_ctxt['system'], flt_ctxt['system_extra']
    return (
        'etree',
        flt_ctxt.ctxt_proceed_xslt(
            in_obj,
            def_first=xslt_params(
                pcscmd_cman=cluster_pcs_flatiron(*sys_pair),
                pcscmd_init_sys=system('init-sys', *sys_pair),
                pcscmd_tomcat_catalina_home=join(
                    '/usr/share',
                    package('tomcat', *sys_pair)
                ),
                pcscmd_named_qual=component_or_state('resource-agents[named]',
                                                     nohitmsg='', *sys_pair),
            )
        )
    )



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Command context, i.e., state distributed along filters chain"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from collections import MutableMapping
from logging import getLogger

from .error import ClufterError
from .utils import isinstanceexcept
from .utils_prog import TweakedDict

log = getLogger(__name__)


class CommandContextError(ClufterError):
    pass


class CommandContextBase(TweakedDict):
    """Object representing command context"""
    def __init__(self, initial=None, parent=None, **kwargs):
        super(CommandContextBase, self).__init__(initial=initial, **kwargs)
        if parent is not None:
            self._parent = parent

    @property
    def anabasis(self):
        """Traverse nested contexts hierarchy upwards"""
        cur = self
        while True:
            yield cur
            if cur is cur._parent:
                break
            cur = cur._parent

    @property
    def parent(self):
        return self._parent

    def __setitem__(self, key, value):
        # XXX value could be also any valid dict constructor argument
        if any(getattr(p, '_notaint', False) for p in self.anabasis):
            raise RuntimeError("Cannot set item in notaint context")
        self._dict[key] = CommandContextBase(initial=value, parent=self) \
                          if isinstanceexcept(value, MutableMapping,
                                                     CommandContextBase) \
                          else value


class CommandContext(CommandContextBase):
    class notaint_context(CommandContextBase.notaint_context):
        def __init__(self, self_outer, exit_off):
            super(self.__class__, self).__init__(self_outer, exit_off)
            self._fc = self_outer['__filter_context__'] \
                       .prevented_taint(exit_off)
        def __enter__(self):
            super(self.__class__, self).__enter__()
            self._fc.__enter__()
        def __exit__(self, *exc):
            self._fc.__exit__()
            super(self.__class__, self).__exit__()

    def __init__(self, *args, **kwargs):
        # filter_context ... where global arguments for filters to be stored
        # filters        ... where filter instance + arguments hybrid is stored
        super(CommandContext, self).__init__(*args, **kwargs)
        # could be cycle up to self if not bypassed
        self['__filter_context__'] = {}  # here we actually want a fallback
        self['__filters__'] = CommandContextBase()

    @staticmethod
    def _wrapping_nested_context(obj):
        class wrapped(CommandContextBase):
            def __getattribute__(self, name):
                if name == 'ctxt_wrapped':
                    ret = obj
                elif name == 'ctxt_set':
                    ret = lambda self, **kwargs: self.update(kwargs)
                elif name.startswith('ctxt_'):
                    # by convention, ctxt_* methods are using second
                    # argument to pass the respective (nested) context
                    ret = obj.__getattribute__(name)
                    if callable(ret):
                        ret = \
                            lambda *args, **kwargs: \
                                obj.__getattribute__(name)(self, *args, **kwargs)
                else:
                    try:
                        if name.startswith('_'):
                            raise KeyError  # dot=index access not for internals
                        ret = self.__getitem__(name)
                    except KeyError:
                        try:
                            ret = super(wrapped, self).__getattribute__(name)
                        except AttributeError:
                            ret = obj.__getattribute__(name)
                return ret

            def __setattribute__(self, name, value):
                obj.__setattribute__(name, value)
        return wrapped

    def ensure_filter(self, flt):
        existing, key = self['__filters__'], flt.__class__.name
        ret = existing.get(key, None)
        if ret is not None:
            assert id(ret.ctxt_wrapped) == id(flt)
        else:
            ret = self._wrapping_nested_context(flt)
            ret = existing[key] = ret(parent=self['__filter_context__'])
        return ret

    def ensure_filters(self, flts):
        return tuple(self.ensure_filter(f) for f in flts)

    def filter(self, which=None):
        if which is not None:
            ret = self['__filters__'][which]
        else:
            ret = self['__filter_context__']
        return ret



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

# following 2nd chance import is to allow direct usage context (testing, etc.)
try:
    from ....filters._2pcscmd import coro2pcscmd
except ValueError:  # Value?
    from ...filters._2pcscmd import coro2pcscmd

needlexml2pcscmd = coro2pcscmd(node='', quorum='', totem='')



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....filters.cib2pcscmd import attrset_xsl
from ....utils_xslt import NL, xslt_is_member

cib2pcscmd = ('''\
    <xsl:for-each select="*[
''' + (
                          xslt_is_member('name()', ('primitive', 'group'))
) + ''']">
''' + (
        verbose_inform('"make ", name(), " a clone: ", @id')
) + '''
        <xsl:value-of select="concat($pcscmd_pcs, 'resource ', name(..))"/>
        <!-- pcs only allows to identify master, not a clone -->
        <xsl:if test="name(..) = 'master'">
            <xsl:value-of select="concat(' ', ../@id)"/>
        </xsl:if>
        <xsl:value-of select="concat(' ', @id)"/>
''' + (
            attrset_xsl("../meta_attributes")
) + '''
        <xsl:value-of select="'%(NL)s'"/>
''' + (
        verbose_ec_test
) + '''
    </xsl:for-each>
''') % dict(
    NL=NL,
)



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from ....filters.cib2pcscmd import attrset_xsl


cib2pcscmd = (
    attrset_xsl("meta_attributes",
                cmd='$pcscmd_pcs, "resource defaults"',
                inform='"set resource defaults"')
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl


ccsflat2cibprelude = '''\
    <!--
        LVM ~ lvm
     -->

    <!-- SHOW-STOPPERS -->

    <xsl:when test="name() = 'lvm'
                    and
                    @lv_name
                    and
                    count(
                        (following-sibling::lvm
                        |../following-sibling::service/lvm)[
                            @vg_name = current()/@vg_name
                            and
                            @lv_name != current()/@lv_name
                        ]
                    ) = 1">
        <xsl:message terminate="yes">
            <xsl:value-of select="concat('Cannot convert lvm when there are',
                                         ' multiple LV bindings for single',
                                         '`', @vg_name, '` VG')"/>
        </xsl:message>
    </xsl:when>

    <!-- STANDARD MODE OF OPERATION -->

    <xsl:when test="name() = 'lvm'">
''' + (
        ResourceSpec('ocf:heartbeat:LVM').xsl_attrs
) + '''
        <xsl:if test="@lv_name">
            <xsl:message>
                <xsl:value-of select="concat('NOTE: LV binding `', @lv_name,
                                             '` unused, it is assumed you have',
                                             ' just a single LV at `', @vg_name,
                                             '` VG')"/>
            </xsl:message>
        </xsl:if>

        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('volgrpname', 'vg_name', req=True)
            +
            rg2hb_xsl('exclusive', 'true', req=abs)
) + '''\
        </instance_attributes>
    </xsl:when>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs-obfuscate command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
from ..protocol import protocols
from ..utils_cman import PATH_CLUSTERCONF


@Command.deco(('ccs-obfuscate-credentials',
                  ('ccs-obfuscate-identifiers')))
def ccs_obfuscate(cmd_ctxt,
                  input=PATH_CLUSTERCONF,
                  output="cluster-obfuscated-"
                         "{ccs-obfuscate-credentials.in.hash}.conf",
                  skip='none'):
    """Obfuscate credentials/IDs in CMAN-based cluster config.

    Either obfuscation pass can be suppressed by skip parameter, by
    default they are performed both in row.

    Following conventions are used for substituted ids/credentials:
    1. identifiers used for crosslinking (referential integrity)
       ought to be converted in a way not violating this integrity
    2. identifiers clearly out of referential integrity (i.e.,
       arbitrary value unrelated to the rest of the XML tree)
       ought to be substituted with strings starting with 'REL-'
    3. credentials ought to be substituted with strings starting
       with 'SECRET-'
    4. overall, any affected item should be substituted with
       capitalized string to visually emphasize the substitution

    Options:
        input   input CMAN-based cluster configuration file
        output  output file with obfuscated credentials/identifiers
        skip    pass to skip (none/ids/creds), neater than --noop
    """
    try:
        skip = ('creds', 'ids').index(skip.lower()) + 1
    except ValueError:
        skip = 0
    if skip == 1:
        cmd_ctxt['filter_noop'].append('ccs-obfuscate-credentials')
    if skip == 2:
        cmd_ctxt['filter_noop'].append('ccs-obfuscate-identifiers')

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        (
            file_proto(output),
        ),
    )



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""pcs-resource-deps filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import XMLFilter
from ..utils import args2sgpl
from ..utils_cib import ResourceSpec
from ..utils_xslt import NL

def map_resource_to_pkgs(spec, *pkgs):
    pkgs = args2sgpl(pkgs)
    return (
        '<xsl:if test="' + ResourceSpec(spec).xsl_attrs_select + '">'
            '<xsl:value-of select="\'' + (NL.join(pkgs)) + NL + '\'"/>'
        '</xsl:if>'
    )


@XMLFilter.deco('cib', 'string-list')
def pcs_resource_deps(flt_ctxt, in_obj):
    """Outputs set of logical dependencies based on configured resources"""
    return ('bytestring', flt_ctxt.ctxt_proceed_xslt(in_obj, textmode=True))



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccs_obfuscate_identifiers = '''\
    <xsl:copy>
        <!-- awkward way to keep both relative attribute ordering
             and ordering QUORUMD-LABEL-{1, 2}; not scaling well -->
        <xsl:variable name="QuorumdLabel"
                      select="@*[
''' + (
            xslt_is_member('name()', ('cman_label',
                                      'label'))
) + ''']"/>
        <xsl:for-each select="@*">
            <xsl:variable name="QuorumdLabelMatch"
                          select="$QuorumdLabel[
                                      name()
                                      =
                                      name(current())
                                  ][1]"/>
            <xsl:choose>
                <xsl:when test="$QuorumdLabelMatch">
                    <xsl:variable name="QuorumdLabelsCurrent"
                                  select="name()"/>
                    <xsl:attribute name="{name()}">
                        <xsl:value-of select="concat(
                            'QUORUMD-LABEL-',
                             count($QuorumdLabelMatch/preceding-sibling::*) + 1
                        )"/>
                    </xsl:attribute>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:copy/>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:for-each>
        <xsl:apply-templates/>
    </xsl:copy>
'''

###

from ....filters.ccs_artefacts import artefact_cond

ccs_artefacts = artefact_cond('@statusfile', kind='F',
                              desc='quorumd status file')



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""stringiter-combine filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from itertools import chain

from ..filter import Filter


def stringiter_combine(flt_ctxt, in_objs):
    """Combine multiple string-iter objects"""
    return (
        'bytestringiter',
        chain(*tuple(o('bytestringiter', protect_safe=True) for o in in_objs))
    )


@Filter.deco(('string-iter', ) * 2, 'string-iter')
def stringiter_combine2(flt_ctxt, in_objs):
    return stringiter_combine(flt_ctxt, in_objs)


@Filter.deco(('string-iter', ) * 3, 'string-iter')
def stringiter_combine3(flt_ctxt, in_objs):
    return stringiter_combine(flt_ctxt, in_objs)


@Filter.deco(('string-iter', ) * 4, 'string-iter')
def stringiter_combine4(flt_ctxt, in_objs):
    return stringiter_combine(flt_ctxt, in_objs)

# ...



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Chains of filters used in *2pcs* commands"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from os import fchmod, fstat
import stat

from ..utils import args2tuple, args2unwrapped, tuplist
from ..utils_func import apply_aggregation_preserving_passing_depth


terminalize = lambda args: \
    apply_aggregation_preserving_passing_depth(
        lambda i, d:
            tuple(a for a in i[:-1] if tuplist(a))
            + tuple([args.pop() if not tuplist(i[-1]) else i[-1]])
    )


def cast_output(chain):
    def cast_output_inner(*args):
        args = list(args)
        args.reverse()
        ret = terminalize(args)(chain)
        return ret
    return cast_output_inner


ccsflat2cibfinal_chain_exec = lambda cont=(): \
    ('ccs-revitalize',
        ('ccsflat2cibprelude',
            ('cibprelude2cibcompact',
                ('cibcompact2cib',
                    (args2unwrapped('cib2cibfinal',
                                    *(cont and args2tuple(cont))))))))
ccsflat2cibfinal_chain = ccsflat2cibfinal_chain_exec()
ccsflat2cibfinal_output = cast_output(ccsflat2cibfinal_chain)

cib2pcscmd_chain_exec = lambda cont=(): \
    ('cib-revitalize',
        ('cib-meld-templates',
            (args2unwrapped('cib2pcscmd',
                            *(cont and args2tuple(cont))))))
cib2pcscmd_chain = cib2pcscmd_chain_exec()
cib2pcscmd_output = cast_output(cib2pcscmd_chain)

ccsflat2pcscmd_chain_exec = lambda cont=(): \
    ccsflat2cibfinal_chain_exec(cib2pcscmd_chain_exec(cont))
#ccsflat2pcscmd_chain = (ccsflat2cibfinal_chain_exec(cib2pcscmd_chain))
ccsflat2pcscmd_chain = ccsflat2pcscmd_chain_exec()
ccsflat2pcscmd_output = cast_output(ccsflat2pcscmd_chain)


def output_set_exec(cmd_ctxt, output_flt):
    """Common post-processing for commands producing scripts, sets exec bits"""
    o = cmd_ctxt.filter(output_flt)['out'].FILE()
    if o.startswith('<') and o.endswith('>'):
        pass  # do not try to manipulate with stdout/stderr
    else:
        try:
            with open(o, 'rb') as f:
                fd = f.fileno()
                fchmod(fd, fstat(fd).st_mode | stat.S_IXUSR | stat.S_IXGRP)
        except IOError:
            from sys import stderr
            svc_output = cmd_ctxt.get('svc_output',
                                      lambda s, **kw: stderr.write(s + '\n'))
            svc_output("Cannot set output file `{0}` executable".format(o),
                       base="error", urgent=True)



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""ccs-artefacts filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import XMLFilter


def artefact(xpath, kind='?', own=False, desc=''):
    return '''\
    <xsl:value-of select="concat('{kind}; {own}; ', {xpath}, '; {desc}', '&#xA;')"/>
'''.format(xpath=xpath, kind=kind, own='1' if own else '0', desc=desc)


def artefact_cond(cond, xpath=None, **kwargs):
    if xpath is None:
        xpath = '.'
    return '''\
    <xsl:for-each select="{cond}">
        {artefact}</xsl:for-each>
'''.format(cond=cond, artefact=artefact(xpath, **kwargs))

def artefact_cond_ra(*args, **kwargs):
    kwargs['desc'] = nested_concat('RA/${name(..)}: ' + kwargs.get('desc'))
    return artefact_cond(*args, **kwargs)


def nested_concat(*args, **kwargs):
    """Make a concat body of argument(s) of strings and ${XSLT expressions}

    Keyword arguments:
        proper_nesting: whether initial and trailing quote is put externally

    Needlessly careless: produces too many extraneous empty strings ('').
    """
    proper_nesting = kwargs.get('proper_nesting', True)
    acc, part = '', ('', '', '}' + ''.join(args))
    while part[2]:
        if part[0]:
            acc += "', '" + part[0]
        xslt = part[2][:part[2].find('}')]
        if xslt:
            acc += "', " + xslt + ", '"
        part = part[2][len(xslt)+1:].partition("${")
    acc += part[0]
    return acc if proper_nesting else acc.join("''")


@XMLFilter.deco('ccs', 'SimpleFormat')
def ccs_artefacts(flt_ctxt, in_obj):
    """Outputs CVS-like file (kind; own; id; description) for CCS artefacts

    Note that "CCS artefacts" refers to any notable file/port/system-wide user,
    etc.  Boolean flag "own" is used to distinguish the assets/something
    provided (exposed) by the cluster stack itself and the artefacts being
    merely consumed.

    Kind can be:
        A   administrative file (mostly synonym to "application-specific
                                 configuration")
        C   command (binary, optionally with full path and arguments)
        D   directory
        F   general file
        I   IP address
        L   log file
        S   sensitive file
        X   executable file (binary or script with shebang)
    (perhaps also:
        G   group
        P   port (optionally suffixed with '[{tcp,udp}]'?)
        U   user
    )
    """
    return ('bytestring', flt_ctxt.ctxt_proceed_xslt(in_obj, textmode=True))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Format representing a list of strings (new-line delimited)"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..format import SimpleFormat
from ..protocol import Protocol
from ..utils_2to3 import bytes_enc, str_enc


class string_iter(SimpleFormat):
    native_protocol = BYTESTRINGITER = Protocol('bytestringiter')
    BYTESTRING = SimpleFormat.BYTESTRING

    @SimpleFormat.producing(BYTESTRING, chained=True)
    def get_bytestring(self, *iodecl):
        """Return command as canonical single string"""
        # chained fallback
        return (
            # add a trailing new-line as per conventions
            # XXX should there be a knob for that?
            bytes_enc(
                '\n'.join(str_enc(s, 'utf-8') for s
                          in self.BYTESTRINGITER(protect_safe=True) if s)
                + '\n',
                'utf-8'
            )
        )

    @SimpleFormat.producing(BYTESTRINGITER, protect=True)
    def get_bytestringiter(self, *iodecl):
        ret = (bytes_enc(s.strip(), 'utf-8')
               for s in (str_enc(self.BYTESTRING(), 'utf-8').rstrip('\n'))
                                                            .split('\n'))
        return ret


class string_list(string_iter):
    native_protocol = BYTESTRINGLIST = Protocol('bytestringlist')

    @SimpleFormat.producing(BYTESTRINGLIST, protect=True)
    def get_bytestringlist(self, *iodecl):
        return list(self.BYTESTRINGITER(protect_safe=True))


class string_set(string_iter):
    native_protocol = BYTESTRINGSET = Protocol('bytestringset')

    @SimpleFormat.producing(BYTESTRINGSET, protect=True)
    def get_bytestringset(self, *iodecl):
        return set(self.BYTESTRINGITER(protect_self=True))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccs2needlexml_attrs = (
    'consensus',
    'fail_recv_const',
    'join',
    'keyfile',
    'max_messages',
    'miss_count_const',
    'netmtu',
    'rrp_mode',
    'rrp_problem_count_threshold',
    'secauth',
    'seqno_unchanged_const',
    'token',
    'token_retransmits_before_loss_const',
    'window_size',
)

ccs2needlexml = '''\
    <xsl:copy-of select="@*[
''' + (
        xslt_is_member('name()', ccs2needlexml_attrs)
) + ''']"/>
    <clufter:descent at="interface"/>
'''

###

from ....filters.ccs_artefacts import artefact_cond

ccs_artefacts = artefact_cond('@keyfile', kind='F', desc='totem keyfile')

###

ccspcmk2pcscmd_rrp_mode = ('active', 'passive')
# 1:1 mapping of supported params to pcs arguments (not: rrpmode)
ccspcmk2pcscmd_supported = (
    'consensus',
    'join',
    'token',
    'fail_recv_const',
    'miss_count_const',
)

ccspcmk2pcscmd = '''\
    <xsl:for-each select="@*[
''' + (
        xslt_is_member('name()', ccspcmk2pcscmd_supported)
) + ''']">
        <xsl:if test=".">
            <xsl:value-of select="concat(' --', name(), ' ', .)"/>
        </xsl:if>
    </xsl:for-each>

    <!-- rrpmode -->
    <xsl:if test="
''' + (
    xslt_is_member('@rrp_mode', ccspcmk2pcscmd_rrp_mode)
) + '''">
        <xsl:value-of select="concat(' --rrpmode ', @rrp_mode)"/>
    </xsl:if>

    <clufter:descent-mix at="interface"/>

    <!-- corosync encryption _disabling_ (not possible) -->
    <xsl:if test="@secauth = 'off'">
        <xsl:message>
            <xsl:value-of select="concat('WARNING: no encryption requested,',
                                         ' but current pcs not capable to',
                                         ' disable it for CMAN')"/>
        </xsl:message>
    </xsl:if>
'''



# -*- coding: UTF-8 -*-
# Copyright 2018 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

###

from ....filters._2pcscmd import verbose_ec_test, verbose_inform
from ....utils_xslt import NL, xslt_is_member

needleqdevicexml2pcscmd_attrs = (
    # model has a special role here
    'sync_timeout',
    'timeout',
    'votes',  # extra handling (e.g. set to 1 for ffsplit algorithm to work)
)
needleqdevicexml2pcscmd_heuristics_attrs = (
    # exec_NAME needs to be handled separately
    'interval',
    'mode',
    'sync_timeout',
    'timeout',
)
needleqdevicexml2pcscmd = ('''\
    <!-- "pcs quorum device add" only supported with certain newer
         versions of pcs, support for qdevice itself in corosync
         ditto -->
    <xsl:choose>
        <xsl:when test="$pcscmd_extra_qdevice">
            <xsl:variable name="ModelOpts">
                <xsl:choose>

                    <!-- make sure there is at least a single "when" -->
                    <xsl:when test="false()"/>

                    <!-- the nested snippets should be guarded with
                         xsl:when test="@model = 'foo'" -->
                    <clufter:descent-mix at="*"/>

                    <xsl:otherwise>
                        <xsl:value-of select="'clufter:UNKNOWN'"/>
                    </xsl:otherwise>
                </xsl:choose>
            </xsl:variable>
            <xsl:choose>
                <xsl:when test="not(starts-with($ModelOpts, 'clufter:'))">
''' + (
                    verbose_inform('"add quorum device: ", @model, " model"')
) + '''
                    <xsl:value-of select="'pcs quorum device add'"/>

                    <xsl:for-each select="@*[
''' + (
                        xslt_is_member('name()', needleqdevicexml2pcscmd_attrs)
) + ''']">
                        <xsl:choose>
                            <xsl:when test="name() = 'votes'">
                                <xsl:message>
                                    <xsl:value-of select="concat('NOTE: `quorum.device.votes`',
                                                                 ' value `', ., '` specified,',
                                                                 ' but current pcs manages it',
                                                                 ' solely on its own')"/>
                                </xsl:message>
                            </xsl:when>
                            <xsl:otherwise>
                                <xsl:value-of select="concat(' ', name(), '=', .)"/>
                            </xsl:otherwise>
                        </xsl:choose>
                    </xsl:for-each>

                    <xsl:value-of select="concat(' model ', @model, $ModelOpts)"/>

                    <xsl:choose>
                        <xsl:when test="$pcscmd_extra_qdevice_heuristics
                                        and
                                        heuristics/@*">
                            <xsl:value-of select="' heuristics'"/>
                            <xsl:for-each select="heuristics/@*[
''' + (
                                xslt_is_member('name()', needleqdevicexml2pcscmd_heuristics_attrs)
) + ''']">
                                <xsl:value-of select="concat(' ', name(), '=', .)"/>
                            </xsl:for-each>
                            <xsl:if test="not(heuristics/@*[
                                              starts-with(name(), 'exec_')
                                          ])">
                                <xsl:message>%(qheuristics_disabled_msg)s</xsl:message>
                            </xsl:if>
                            <xsl:for-each select="heuristics/@*[
                                                      starts-with(name(), 'exec_')
                                                  ]">
                                <xsl:value-of select='concat(" &apos;", name(), "=", .,
                                                             "&apos;")'/>
                            </xsl:for-each>
                        </xsl:when>
                        <xsl:when test="$pcscmd_extra_qdevice_heuristics">
                            <!-- no parameters, nothing to do -->
                        </xsl:when>
                        <xsl:otherwise>
                            <xsl:message>%(qheuristics_msg)s</xsl:message>
                        </xsl:otherwise>
                    </xsl:choose>

                    <xsl:value-of select="'%(NL)s'"/>
''' + (
                    verbose_ec_test
) + '''
                </xsl:when>
                <xsl:when test="$ModelOpts = 'clufter:UNKNOWN'">
                    <xsl:message>%(note_unhandled)s</xsl:message>
                </xsl:when>
                <xsl:when test="$ModelOpts = 'clufter:UNSUPPORTED'">
                    <xsl:message>%(qmodel_msg)s</xsl:message>
                </xsl:when>
                <xsl:otherwise>
                    <xsl:message>WARNING: internal error</xsl:message>
                </xsl:otherwise>
            </xsl:choose>
        </xsl:when>
        <xsl:otherwise>
            <xsl:message>%(qdevice_msg)s</xsl:message>
        </xsl:otherwise>
    </xsl:choose>
''') % dict(
    NL=NL,
    note_unhandled='''<xsl:value-of select="concat('WARNING: quorum device model `',
                                                   @model,
                                                   '` is currently unhandled',
                                                   ' in the conversion')"/>''',
    qdevice_msg="WARNING: target corosync/pcs versions do not support qdevice,"
                " hence omitted",
    qheuristics_msg="WARNING: target corosync/pcs versions do not support"
                    " heuristics, hence omitted",
    qheuristics_disabled_msg="WARNING: no heuristic commands specified,"
                             " heuristics will not be enabled by corosync",
    qmodel_msg='''<xsl:value-of select="concat('WARNING: target corosync/pcs',
                                               ' versions do not support quorum',
                                               ' device model `',
                                               @model,
                                               '`, hence omitted')"/>''',
)



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cib2pcscmd command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
try:
    from ..defaults import SHELL_POSIX
except ImportError:
    SHELL_POSIX = ''
from ..filter import XMLFilter
from ..protocol import protocols
from ..utils_cib import PATH_CIB
from ._chains_pcs import cib2pcscmd_chain_exec  #, cib2pcscmd_output


@Command.deco(('cmd-annotate',
                  ('stringiter-combine2',
                      ('cmd-wrap'))),
              (cib2pcscmd_chain_exec(
                  ('stringiter-combine2'  # , ('cmd-wrap' ...
                   ))))
def cib2pcscmd(cmd_ctxt,
               input=PATH_CIB,
               output="-",
               force=False,
               noauth=False,
               silent=False,
               tmp_cib="{cib2pcscmd.defs[pcscmd_tmpcib]}",
               dry_run=False,
               set_exec=False,
               text_width='0',
               _common=XMLFilter.command_common):
    """CIB -> equivalent in pcs commands

    Options:
        input       input proper Pacemaker cluster configuration file (CIB)
        output      pcs commands to reinstate the cluster per the inputs
        force       may the force be with emitted pcs commands
        noauth      skip authentication step (OK if already set up)
        silent      do not track the progress along the steps execution (echoes)
        tmp_cib     file to accumulate the changes (empty ~ direct push, avoid!)
        dry_run     omit intrusive commands (TMP_CIB reset if empty)
        set_exec    make the output file executable (not recommended)
        text_width  for commands rewrapping (0/-1/neg. ~ auto/disable/hi-limit)
    """
    cmd_ctxt['pcscmd_force'] = force
    cmd_ctxt['pcscmd_noauth'] = noauth
    cmd_ctxt['pcscmd_verbose'] = not(silent)
    cmd_ctxt['pcscmd_tmpcib'] = tmp_cib
    cmd_ctxt['pcscmd_dryrun'] = dry_run
    cmd_ctxt['text_width'] = text_width

    cmd_ctxt['annotate_shell'] = SHELL_POSIX
    # XXX possibility to disable cib-meld-templates

    void_proto = protocols.plugins['void'].ensure_proto
    file_proto = protocols.plugins['file'].ensure_proto
    yield (
        (
            void_proto(),
            (
                    (
                        file_proto(output),
                    ),
            ),
        ), (
            file_proto(input),
            # already tracked
            #cib2pcscmd_output(
            #        (
            #            file_proto(output),
            #        ),
            #),
        ),
    )
    # post-processing (make resulting file optionally executable)
    if set_exec:
        output_set_exec(cmd_ctxt, 'cmd-wrap')



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Program-specific commons"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

import logging
from collections import Mapping, MutableMapping, MutableSequence, MutableSet
from functools import reduce
from optparse import Option
from os import environ, fdopen, isatty, pathsep
from os.path import abspath, dirname, samefile, \
                    isabs as path_isabs, \
                    isfile as path_isfile, \
                    join as path_join
from re import compile as re_compile
from subprocess import Popen

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from sys import stderr, stdin, stdout

from . import package_name
from .error import ClufterError
from .utils import areinstances, \
                   filterdict_pop, \
                   func_defaults_varnames, \
                   isinstanceexcept, \
                   selfaware, \
                   tuplist
from .utils_2to3 import basestring, foreach_u, iter_items, xrange
from .utils_func import apply_split

# do not alias it other way around to avoid accidental "file(f, 'w')"
try:
    IOBase = file
except NameError:  # PY3 forward compability
    from io import IOBase

log = logging.getLogger(__name__)


#
# generics
#

mutables = (MutableMapping, MutableSequence, MutableSet)

class TweakedDict(MutableMapping):
    """Object representing command context

       Note: not suitable for wrapping OrderedDict or the like as that
             extra property will get disregarded (+ warning emitted)."""

    class notaint_context(object):
        def __init__(self, self_outer, exit_off):
            self._exit_off = exit_off
            self._self_outer = self_outer
        def __enter__(self):
            self._exit_off |= not self._self_outer._notaint
            self._self_outer._notaint = True
        def __exit__(self, *exc):
            self._self_outer._notaint = not self._exit_off

    def __init__(self, initial=None, bypass=False, notaint=False):
        self._parent = self
        self._notaint = True
        if areinstances(initial, self):
            assert initial._parent is initial
            self._dict = initial._dict  # trust dict to have expected props
            notaint = initial._notaint
        else:
            self._dict = {}
            if initial is not None:
                if not isinstance(initial, Mapping):
                    initial = dict(initial)
                elif not isinstance(initial, MutableMapping):
                    # silently? follow the immutability
                    notaint = True
                    bypass = True
                elif isinstance(initial, OrderedDict):
                    log.warn("Possibly dropping 'ordered' property of dict")
                if bypass or notaint:
                    self._dict = initial
                if not bypass:
                    # full examination
                    self._notaint = False  # temporarily need to allow
                    foreach_u(lambda k, v: self.__setitem__(k, v),
                              iter_items(initial))
        self._notaint = notaint

    def __delitem__(self, key):
        if any(getattr(p, '_notaint', False) for p in self.anabasis):
            raise RuntimeError("Cannot del item in notaint context")
        del self._dict[key]

    def __getitem__(self, key):
        # any notainting parent incl. self is an authority for us
        try:
            ret = self._dict[key]
        except KeyError:
            if self._parent is self:
                raise
            ret = self._parent[key]
        if (isinstanceexcept(ret, mutables, TweakedDict)
            and any(getattr(p, '_notaint', False) for p in self.anabasis)):
            ret = type(ret)(ret)
        return ret

    @property
    def anabasis(self):
        """Traverse nested contexts hierarchy upwards"""
        return (self, )

    def setdefault(self, key, *args, **kwargs):
        """Allows implicit arrangements to be bypassed via `bypass` flag"""
        assert len(args) < 2
        bypass = kwargs.get('bypass', False)
        if bypass:  # for when adding MutableMapping that should be untouched
            return self._dict.setdefault(key, *args)
        try:
            return self.__getitem__(key)
        except KeyError:
            if not args:
                raise
            self.__setitem__(key, *args)
            return args[0]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return "<{0}: {1}>".format(repr(self.__class__), repr(self._dict))

    def __setitem__(self, key, value):
        # XXX value could be also any valid dict constructor argument
        if any(getattr(p, '_notaint', False) for p in self.anabasis):
            raise RuntimeError("Cannot set item in notaint context")
        self._dict[key] = value

    def prevented_taint(self, exit_off=False):
        """Context manager to safely yield underlying dicts while applied"""
        return self.notaint_context(self, exit_off)

ProtectedDict = lambda track: TweakedDict(track, notaint=True, bypass=True)


#
# command-line options related
#

# dashes recommended by 5 of 4 terminal fanboys
# NB underscores -> dashes swap is idempotent in regards to the member
#    name of where optparse stores respective option value, but here
#    the list of possible operations to preserve such property stops!
#    + also to preserve reversibility/bijection, original ought to be
#      free of the substituting character
cli_decor = lambda x: x.replace('_', '-')
cli_undecor = lambda x: x.replace('-', '_')

# prioritize consonants, deprioritize vowels (except for the first letter
# overall), which seems to be widely adopted technique for selecting short
# options based on their long counterparts ;)
# XXX could eventually also resort to using upper-case chars
longopt_letters_reprio = \
    lambda longopt: \
        (lambda lo:
            lo[0] + ''.join(sorted(lo[1:],
                                   key=lambda x: int(x.lower() in 'aeiouy')))
        )(tuple(c for c in longopt if c.isalpha()))

class ExpertOption(Option):
    pass

make_option = \
    lambda *a, **kw: \
        (ExpertOption if kw.pop('expert', False) else Option)(*a, **kw)

# extrapolate make_option to specifically-encoded "plural"
make_options = lambda opt_decl: [make_option(*a, **kw) for a, kw in opt_decl]


def set_logging(opts):
    """Apply logging options as per `opts' to the live logging configuration"""
    rootlog = logging.getLogger()
    last_hdlr = rootlog.handlers.pop()
    if isinstance(last_hdlr, logging.FileHandler if opts.logfile
                             else logging.StreamHandler) \
      and (samefile(opts.logfile, last_hdlr.baseFilename)
           if opts.logfile else last_hdlr.stream is stderr):
        hdlr = last_hdlr
    else:
        hdlr = logging.FileHandler(opts.logfile) if opts.logfile \
               else logging.StreamHandler()
        hdlr.setFormatter(last_hdlr.formatter)
    rootlog.addHandler(hdlr)
    rootlog.setLevel(logging.getLevelName(opts.loglevel))


def PopenLog(cmd, *args, **kwargs):
    log.debug("Running: {0}".format(' '.join(cmd)))
    return Popen(cmd, *args, **kwargs)


class OneoffWrappedStdinPopen(object):
    """Singleton to watch for atmost one use of stdin in Popen context"""
    def __init__(self):
        self._used = False

    def __call__(self, args, **kwargs):
        if not 'stdin' in kwargs and '-' in args:
            if self._used:
                raise ClufterError(self, 'repeated use detected')
            kwargs['stdin'] = stdin
            # only the first '-' substituted
            args[args.index('-')] = '/dev/stdin'
            self._used |= True
        return PopenLog(args, **kwargs)

OneoffWrappedStdinPopen = OneoffWrappedStdinPopen()


#
# misc
#

# NB: distutils.spawn.find_executable + shutil.which (Python 3.3+)
def which(name, single='', *paths, **redefine_check):
    """Mimic `which' UNIX utility

    Both `single` and `paths` denotes paths to be tried for `name` lookup,
    which includes a decision from the perspective of a filter function given
    by `check` keyword argument (default: a test if the file exists at all).

    What is special about `single` is that it can be defined as either
    plain one-path string, PATH-like-separated list of paths, or
    an iterable.  It is decomposed into a list of paths and then
    `paths` items appended.

    Apparently, when `name` is provided as an absolute path, only
    the `check` (i.e., no lookup) phase is performed.

    If nothing matching the criteria is found, `None` is returned,
    nominal path as per `name` (and `check` for that matter) otherwise.

    Empty string instructs code to use environment's PATH instead.
    """
    check, expand_path = redefine_check.pop('check', path_isfile), True
    if path_isabs(name):
        where, expand_path = [''], False
    else:
        where = list(single.split(pathsep) if not tuplist(single) else single)
        where.extend(paths)
        where.reverse()
    while where:
        p = where.pop()
        if not p:
            if expand_path:
                where.extend(reversed(environ.get('PATH', '').split(pathsep)))
                expand_path = False
                continue
            elif where:
                break  # degenerated case: multiple ''
        p = path_join(abspath(p), name)
        if check(p):
            return p
    else:
        return None


dirname_x = lambda p, c=1: reduce(lambda x, y: dirname(x), xrange(c), p)


@selfaware
def defer_common(me, fnc, skip=0):
    """Use when you have a func with common initial kwargs consumption

    Calling this on another function will, upon invocation, first look
    at default argument to `_common` keyword parameter, and if present,
    will consider it as a function to be run before the one passed as
    an argument; aside from returning a function decorated in a stated
    way, it will also return dict of function paramater defaults (joint
    for both original and '_common' function) and a list of joint
    function arguments.

    In other words, it allows DRY principle for common shared initial
    kwargs "consumer" function on top of (actually prior to) the function
    being passed in as an argument.
    """
    fnc_defaults, fnc_varnames = func_defaults_varnames(fnc, skip=skip)
    common = fnc_defaults.pop('_common', None)
    if not common:
        wrapfnc = fnc
    else:
        common_defaults, common_varnames, common = me(common, skip=skip)
        fnc_defaults.update(common_defaults)
        fnc_varnames = list(fnc_varnames)
        fnc_varnames.remove('_common')  # but we could rely on last argument
        fnc_varnames = tuple(fnc_varnames) + common_varnames

        def wrapfnc(cmd_ctxt, **kwargs):
            common(cmd_ctxt, **filterdict_pop(kwargs, *common_varnames))
            kwargs.pop('_common', None)
            return fnc(cmd_ctxt, **kwargs)
        wrapfnc.__doc__ = fnc.__doc__ + common.__doc__
    return fnc_defaults, fnc_varnames, wrapfnc


def getenv_namespaced(varname, value=None, namespace=package_name().upper()):
    """Obtain value of environment variable prefixed with `namespace + '_'`"""
    return environ.get('_'.join((namespace, varname)), value)


def setenv_namespaced(varname, value=None, namespace=package_name().upper()):
    """Set value of environment variable prefixed with `namespace + '_'`

    Variable is removed (i.e. undefined) if `value` is `None`."""
    if value is None:
        try:
            del environ['_'.join((namespace, varname))]
        except KeyError:
            pass
    else:
        environ['_'.join((namespace, varname))] = value


# cf. https://github.com/karelzak/util-linux/blob/master/lib/colors.c#L107
#     https://github.com/karelzak/util-linux/blob/master/include/colors.h#L14
class FancyOutput(object):
    """
    Mean to produce more fancy output based on a simple tagging

    One can tag particular disjoint parts of the message like this:

        This is |error:error| and this is |warning:warning|.

    which produces

        This is error and this is warning.

    but on a terminal, both annotated words will be colorized as specified.
    Currently, it is hard-coded with a possibility for overrides through `cfg`
    collective keyword argument;  in the future terminal-colors.d scheme
    as introduced by util-linux package, see TERMINAL_COLORS.D(5),
    will be leveraged.

    Usage is simple, instantiate an object of this class, then pass it
    around your code and call it with a message to be produced (possibly
    using the introduced tagging) as a parameter.

    Other per-message-emit, keyword, parameters:
        base    additionally wrap the whole message with this default markup
                (only markup is taken into account as such)
        urgent  produce the message regardless of `quiet` parameter
                in the constructor
    """
    logic_colors = (
        'error',
        'header',
        'highlight',
        'note',
        'subheader',
        'warning',

        'pcscmd_comment',
        'pcscmd_file',
        'pcscmd_pcs',
        'pcscmd_subcmd',
        'pcscmd_metaword',
    )
    re_color = re_compile(
        '\|(?P<logic>' + '|'.join(logic_colors) + '):(?P<msg>[^\|]*)\|'
    )

    colors = dict(
        black        = "\033[30m",
        blue         = "\033[34m",
        brown        = "\033[33m",
        cyan         = "\033[36m",
        darkgray     = "\033[1;30m",
        gray         = "\033[37m",
        green        = "\033[32m",
        lightblue    = "\033[1;34m",
        lightcyan    = "\033[1;36m",
        lightgray    = "\033[37m",
        lightgreen   = "\033[1;32m",
        lightmagenta = "\033[1;35m",
        lightred     = "\033[1;31m",
        magenta      = "\033[35m",
        red          = "\033[31m",
        yellow       = "\033[1;33m",

        restore      = '\033[0m',
    )

    table = dict(
        error     = 'lightred',
        header    = 'magenta',
        highlight = 'green',
        note      = 'brown',
        subheader = 'blue',
        warning   = 'red',

        pcscmd_comment  = 'brown',
        pcscmd_file     = 'magenta',
        pcscmd_pcs      = 'blue',
        pcscmd_subcmd   = 'green',
        pcscmd_metaword = 'cyan',
    )

    @classmethod
    def get_color(cls, spec):
        return cls.colors.get(spec, spec if spec.startswith('\033[') or not spec
                                    else spec.join(('\033[', 'm')))

    @classmethod
    def normalized(cls, s):
        assert isinstance(s, basestring)
        return ''.join(map(
            lambda x: x.rsplit('\033[', 1)[0],
            apply_split(
                s,
                lambda i, _, a: i == 'm' and a
                                         and a[-1].rstrip('0123456789;')[-2:]
                                             == '\033['
            )
        ))

    @classmethod
    def len_normalized(cls, s):
        return len(cls.normalized(s))

    # TODO use /etc/terminal-colors.d/clufter.{enable,disable,scheme}
    def __init__(self, f=stdout, recheck=False, color=None, quiet=False,
                 prefix='', **cfg):
        if not isinstance(f, IOBase):
            f = fdopen(f, "a")
        self._f = f
        self._quiet = quiet
        self._prefix = prefix
        self._table = self.table.copy().update(cfg)
        if color is not None:
            recheck = False
        else:
            color = isatty(f.fileno()) if hasattr(f, 'fileno') else False
        self._handle = self.handle_recheck if recheck else \
                       self.handle_color if color else self.handle_std

    def __call__(self, s, **kwargs):
        if self._quiet and not kwargs.pop('urgent', False):
            return
        if self._prefix:
            prefix = self._prefix
            if 'prefix_arg' in kwargs:
                prefix = prefix.format(kwargs['prefix_arg'])
            s = prefix + s
        self._handle(s, **kwargs)
        self._f.flush()

    def handle_recheck(self, s, **kwargs):
        (self.handle_color if isatty(self._f) else self.handle_std)(s, **kwargs)

    def handle_std(self, s, **kwargs):
        self._f.write(self.re_color.sub(lambda m: m.group('msg'), s) + '\n')

    def handle_color(self, s, base=None, **kwargs):
        end = self.colors['restore']
        flip = end + self.get_color(self.table.get(base, ''))
        self._f.write(
            flip
            + self.re_color.sub(
                lambda m:
                    end
                    + self.get_color(self.table.get(m.group('logic'), ''))
                    + m.group('msg')
                    + flip,
                s
            )
            + end
            + '\n'
        )


def docformat(*args, **kwargs):
    """Workaround unability to set __doc__ as an evaluated (subst) expression"""
    def deco(fnc):
        if hasattr(fnc, '__doc__'):
            try:
                fnc.__doc__ = fnc.__doc__.format(*args, **kwargs)
            except IndexError:
                pass
        return fnc
    return deco


def namever_partition(s):
    """For `.*-*[0-9]*(.[0-9]*)*`, return tuple (prefix, v1, ..., vX)

    For instance, 'pacemaker-2.10' -> ('pacemaker', 2, 10)."""
    name = s.rstrip("0123456789.")
    return ((name.rstrip('-'), )
            + tuple(int(i) for i in s[len(name):].split('.') if i))



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

ccs2needlexml = '''\
    <nodelist>
        <clufter:descent at="clusternode"/>
    </nodelist>
'''

ccsflat2cibprelude = '''\
    <nodes>
        <clufter:descent at="clusternode"/>
    </nodes>
'''



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_xslt import xslt_is_member

###

ccs2needlexml_attrs = (
    # 'expected_votes',  # see below
    'two_node',
)

ccs2needlexml = '''\
    <xsl:copy-of select="@*[
''' + (
        xslt_is_member('name()', ccs2needlexml_attrs)
) + ''']"/>
    <xsl:if test="@expected_votes">
        <xsl:message>
            <xsl:value-of select="concat('WARNING: intentionally omitting',
                                         ' `expected_votes` parameter as it',
                                         ' may be based on assumptions',
                                         ' no longer true in post-conversion',
                                         ' environment')"/>
        </xsl:message>
    </xsl:if>
'''

###

from ....filters.ccs_artefacts import artefact_cond

ccs_artefacts = artefact_cond('@keyfile', kind='F', desc='CMAN keyfile')

###

# NOTE pcs doesn't support udpb directly via --transport (unless forced),
#      one has to specify --broadcast0 parameter instead and transport
#      will get silently switched to udpb

ccspcmk2pcscmd_transports = ('udp', 'udpu')

ccspcmk2pcscmd = '''\
    <xsl:if test="
''' + (
    xslt_is_member('@transport', ccspcmk2pcscmd_transports)
) + '''">
        <xsl:value-of select="concat(' --transport ', @transport)"/>
    </xsl:if>

    <clufter:descent-mix at="multicast"/>
    <clufter:descent-mix at="altmulticast"/>
'''



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)


###

ccspcmk2pcscmd = '''\
    <xsl:if test="@addr">
        <xsl:value-of select="concat(' --mcast0 ', @addr)"/>
    </xsl:if>
    <xsl:if test="@port">
        <xsl:value-of select="concat(' --mcastport0 ', @port)"/>
    </xsl:if>
    <xsl:if test="@ttl">
        <xsl:value-of select="concat(' --ttl0 ', @ttl)"/>
    </xsl:if>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from ....utils_2to3 import execfile
from ....utils_prog import dirname_x

from logging import getLogger
log = getLogger(__name__)

# XXX a bit dirty DRY approach
from os.path import exists, join
use = dirname_x(__file__, 6)
use = join(use, 'fencedevices', 'fencedevice')
use = use + '.py' if exists(use + '.py') else join(use, '__init__.py')
myglobals = {'__package__': __package__}
try:
    execfile(use, myglobals)
except IOError:
    log.error("Unable to refer to `{0}' file".format(use))
else:
    ccs_obfuscate_credentials = myglobals['ccs_obfuscate_credentials']
    # XXX might be in tighter bind to particular fencedevice-s
    ccs_artefacts = myglobals['ccs_artefacts_common_params']



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cib-revitalize command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
from ..protocol import protocols
from ..utils_cib import PATH_CIB


@Command.deco('cib-revitalize')
def cib_revitalize(cmd_ctxt,
                   input=PATH_CIB,
                   output="cib-revitalized-{cib-revitalize.in.hash}.xml"):
    """Migrate deprecated configuration items in CIB

    Options:
        input   input proper Pacemaker cluster configuration file (CIB)
        output  output file with "revitalized" content
    """
    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        file_proto(output),
    )



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from .... import package_name

ccsflat2cibprelude = ('''\
    <template id="FAILOVERDOMAIN-{@name}"
              class="ocf"
              provider="%(package_name)s"
              type="temporary-failoverdomain">
        <meta_attributes id="FAILOVERDOMAIN-{@name}-META">
            <nvpair id="FAILOVERDOMAIN-{@name}-META-ordered"
                    name="ordered"
                    value="0">
                <xsl:if test="@ordered != ''
                              and
                              contains('123456789',
                                        substring(@ordered, 1, 1))">
                    <xsl:attribute name="value">
                        <xsl:value-of select="'1'"/>
                    </xsl:attribute>
                </xsl:if>
            </nvpair>
            <nvpair id="FAILOVERDOMAIN-{@name}-META-restricted"
                    name="restricted"
                    value="0">
                <xsl:if test="@restricted != ''
                              and
                              contains('123456789',
                                        substring(@restricted, 1, 1))">
                    <xsl:attribute name="value">
                        <xsl:value-of select="'1'"/>
                    </xsl:attribute>
                </xsl:if>
            </nvpair>
            <nvpair id="FAILOVERDOMAIN-{@name}-META-nofailback"
                    name="nofailback"
                    value="0">
                <xsl:if test="@nofailback != ''
                              and contains('123456789',
                                           substring(@nofailback, 1, 1))">
                    <xsl:attribute name="value">
                        <xsl:value-of select="'1'"/>
                    </xsl:attribute>
                </xsl:if>
            </nvpair>
            <xsl:for-each select="failoverdomainnode">
                <nvpair id="FAILOVERDOMAIN-{../@name}-META-failoverdomainnode-{@name}"
                        name="failoverdomainnode-{@priority}"
                        value="{@name}"/>
            </xsl:for-each>
        </meta_attributes>
    </template>
''') % dict(package_name=package_name())

###

ccs_revitalize = '''\
    <!-- omit failoverdomainnodes with repeated name
         rgmanager/src/daemons/fo_domain.c:fod_get_node:error #30
     -->
    <xsl:template match="failoverdomainnode[
                             preceding-sibling::failoverdomainnode[
                                 @name = current()/@name
                             ]
                         ]">
        <xsl:message>
            <xsl:value-of select="concat('WARNING: omitting failoverdomainnode',
                                         ' with repeated name (', @name, ')')"/>
        </xsl:message>
    </xsl:template>
'''



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""cmd-wrap command"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..command import Command
from ..protocol import protocols


@Command.deco('cmd-wrap')
def cmd_wrap(cmd_ctxt,
             input="-",
             output="-",
             text_width='0'):
    """Wrap long lines carrying (fairly) convoluted shell commands

    Options:
        input       lines carrying moderately convoluted shell commands
        output      (hopefully) the same content fitting the text width
        text_width  for commands rewrapping (0/-1/neg. ~ auto/disable/hi-limit)
    """
    cmd_ctxt['text_width'] = text_width

    file_proto = protocols.plugins['file'].ensure_proto
    return (
        file_proto(input),
        file_proto(output),
    )



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)

from __future__ import print_function

"""Setup script/data"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

try:
    from setuptools import setup, Extension
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, Extension

from collections import Callable
from glob import glob
from os import chdir, getcwd, getenv, sep, walk
from os.path import (join as path_join, realpath as path_real,
                     basename as path_basename, dirname as path_dirname,
                     abspath as path_abs, normpath as path_norm,
                     isabs as path_isabs, isdir as path_isdir,
                     isfile as path_isfile, splitext as path_splitext)
from shutil import copy, copystat
from sys import modules as sys_modules, path as sys_path, prefix as sys_prefix

from distutils.cmd import Command
from distutils.errors import DistutilsSetupError
from distutils.command.build import build
from distutils.command.build_ext import build_ext
from distutils.command.install_data import install_data
from setuptools.command.develop import develop as setuptools_develop
# otherwise fails on
#   error: option --single-version-externally-managed not recognized...managed
# with default pip install
from setuptools.command.install import install as setuptools_install

# Python 3 compatibility
from sys import version_info
if version_info[0] >= 3:
    str_enc = lambda s, encoding='ascii': str(s, encoding)
else:
    str_enc = lambda s, *args: str(s)
try:
    basestring = basestring
except NameError:
    basestring = str

# bail out if any code is not valid (http://stackoverflow.com/a/2240549)
import py_compile
orig_py_compile = py_compile.compile
def doraise_py_compile(file, cfile=None, dfile=None, doraise=False):
    orig_py_compile(file, cfile=cfile, dfile=dfile, doraise=True)
py_compile.compile = doraise_py_compile

PREFER_GITHUB = True
PREFER_FORGE = 'github' if PREFER_GITHUB else 'pagure'  # alternatively: None
DEBUG = getenv("SETUPDEBUG")
DBGPFX = str(__file__)

here = path_abs(path_dirname(path_real(__file__)))
prev_cwd = getcwd()
chdir(here)  # make setup.py possess expected CWD + play better with pip install

#
# Custom machinery extending setuptools/distutils with mechanism
# for parameterization (mainly paths) and even content of these files
#

true_gen = lambda this: True
# XXX copy-paste of utils_func.py
# from functools import reduce
#bifilter = \
#    lambda fnc, seq: \
#        reduce(lambda acc, x: acc[int(not fnc(x))].append(x) or acc,
#               seq, ([], []))

# this is needed to workaround naive approach in recent setuptools
# (https://bitbucket.org/pypa/setuptools/issue/195/no-longer-follows-symbolic)
# which follows symlinks but cannot prune the excluded dirs (perhaps
# declared so as to prevent infinite symlink recursion at the first place!)
# directly at the point of unfolding the directory tree traversal further
def find_packages(where=None, exclude=()):
    ret = []
    if not where:
        where = getcwd()
    excl_set = set(e.strip('*.') for e in exclude)  # rough overapproximation!
    for root, dirs, files in walk(where, followlinks=True):
        pkg_root = root[len(where):].lstrip(sep)
        if '.' in pkg_root:  # avoid *.egg-info and the like/invalid pkg name
            dirs[:] = []
            continue
        pkg_root = pkg_root.replace(sep, '.')
        dirs[:] = [d for d in dirs
                   if d not in excl_set
                   and '.' not in d
                   and path_isfile(path_join(root, d, '__init__.py'))]
        ret.extend('.'.join((pkg_root, d)).lstrip('.') for d in dirs)
    return ret

# this is needed to workaround naive approach in recent setuptools
# (https://bitbucket.org/pypa/setuptools/issue/277/files-in-a-symbol-link-dir)
# which also(!) now follows symlinks (but apparently cannot detect traversal
# infloops!)
from os import curdir
from distutils import filelist
def findall(dir=curdir):
    all_files = []
    for base, dirs, files in walk(dir, followlinks=False):
        if base == curdir or base.startswith(curdir + sep):
            base = base[2:]
        if base:
            files = [path_join(base, f) for f in files]
        all_files.extend(filter(path_isfile, files))
    return all_files
filelist.findall = findall  # fix findall bug in distutils (+ setuptools)


# http://code.activestate.com/recipes/502261-python-distutils-pkg-config/#c1
def pkgconfig(*packages, **kw):
    try:
        from subprocess import check_output
        getoutput = lambda cmd: check_output(cmd, shell=True)
    except ImportError:
        from commands import getoutput
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}
    for token in str_enc(getoutput("pkg-config --libs --cflags {0}"
                                   .format(' '.join(packages)))).split():
        if token[:2] in flag_map:
            kw.setdefault(flag_map.get(token[:2]), []).append(token[2:])
        else:  # throw others to extra_link_args
            kw.setdefault('extra_link_args', []).append(token)
    for k in kw:  # remove duplicates
        kw[k] = list(set(kw[k]))
    return kw

# inspired from speaklater: http://pypi.python.org/pypi/speaklater
class LazyRead(object):
    """Mimic string that in fact is read on-demand from file(s)

    Note: only direct use, str methods and '+' operator supported.
    """
    def __init__(self, filename, postproc=lambda c: c):
        self._filename = path_norm(path_join(here, filename))  # chdir'd, but...
        self._postproc = postproc
        self._content = None
    @property
    def content(self):
        ret = self._content
        if ret is None:
            with open(self._filename) as f:
                ret = self._content = self._postproc(f.read())
        return ret
    def __str__(self):           return str(self.content)
    def __repr__(self):          return repr(self.content)
    def __getattr__(self, what): return getattr(self.content, what)
    def __add__(self, other):    return self.content + other
    def __radd__(self, other):   return other + self.content


class Binary(Extension):
    pass


class build_binary(build_ext):
    @staticmethod
    def prereq(self):
        #return any(isinstance(e, Binary)
        #           for e in self.distribution.ext_modules)
        mine = self.distribution.get_option_dict(build_binary.__name__)
        return bool(mine.get('binaries', ()))

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.binaries = ()

    def finalize_options(self):
        build_ext.finalize_options(self)
        self.extensions = self.binaries
        self.build_lib = path_dirname(self.build_lib)
        #self.extensions, self.distribution.ext_modules = bifilter(
        #    lambda e: isinstance(e, Binary), self.distribution.ext_modules
        #)

    def build_extensions(self):
        compiler = self.compiler
        compiler.link_shared_object = \
            lambda *args, **kwargs: \
                compiler.link_executable(
                    *args, **dict(((k, kwargs[k]) for k in kwargs
                                  if k not in ('build_temp', 'export_symbols',
                                               'libraries')),
                                  libraries=[l for l in
                                             kwargs.get('libraries', ())
                                             if not l.startswith('python')])
                )

        build_ext.build_extensions(self)

    def check_extensions_list(self, extensions):
        if not all(isinstance(i, Binary) for i in extensions):
            raise DistutilsSetupError("Only Binary instances allowed")

    def get_ext_filename(self, ext_name):
        return path_join(*ext_name.split('.'))


class develop(setuptools_develop):
    def install_for_development(self):
        self.reinitialize_command('pkg_prepare', develop=1)
        self.run_command('pkg_prepare')


class install(setuptools_install):
    def finalize_options(self):
        self.single_version_externally_managed = True
        orig_root = self.root  # avoid original install fail on SVEM + not root
        if not orig_root:
            self.root = sep
        setuptools_install.finalize_options(self)
        self.root = orig_root


class install_data(install_data):
    # Change predicate telling whether to run ``install_data''
    # subcommand within ``install'' command to always return true
    prereq = staticmethod(true_gen)


def setup_pkg_prepare(pkg_name, pkg_prepare_options=()):

    class pkg_prepare(Command):
        DEV_SWITCH = 'develop'
        BUILD_DEV_SWITCH = 'build_develop'
        description = ("Prepare specified files, i.e. substitute some values "
                       "(works as a subcommand for ``build'' and ``install'' "
                       "and can also mimic enhanced ``develop'' using %s)"
                       % DEV_SWITCH)
        user_options = [
            (key.replace('_', '-') + "=", None, "specify " + help
             + " for " + pkg_name) for key, help in pkg_prepare_options
        ]
        user_options.append((DEV_SWITCH.replace('_', '-'), None,
                             "mimics ``develop'' command "
                             "(to be used only w/ standalone ``pkg_prepare'')"))
        user_options.append((BUILD_DEV_SWITCH.replace('_', '-'), None,
                             "as --{0} but quit early "
                             "(to be used only w/ standalone ``pkg_prepare'')"
                             .format(DEV_SWITCH)))
        boolean_options = (DEV_SWITCH, BUILD_DEV_SWITCH)
        needs_any_opt = ('package_data', 'data_files',
                         'buildonly_files', 'built_files')
        needs_always_opts = ('pkg_params', )

        @classmethod
        def inject_cmdclass(cls, *new_classes, **orig_classes):
            cmdclass = {}
            for name in orig_classes:
                cmdclass[name], new = cls._inject(name, orig_classes[name])
                new_classes += new
            for c in (cls, ) + new_classes:
                cmdclass[c.__name__] = c
            return cmdclass

        @classmethod
        def _inject(cls, name, orig_cls):
            sub_cmds = (cls, )
            if hasattr(orig_cls, '__iter__'):
                orig_cls, sub_cmds = (lambda c, *s: (c, sub_cmds + s))(*orig_cls)
            sub_cmd_names = set(s.__name__ for s in sub_cmds)
            orig_sub_commands = [s for s in orig_cls.sub_commands
                                 if s[0] not in sub_cmd_names]

            class new_cls(orig_cls):
                sub_commands = [(sub_cmd.__name__, getattr(sub_cmd, 'prereq',
                                                           true_gen))
                                for sub_cmd in sub_cmds] + orig_sub_commands
                description = (orig_cls.description + " (modified: "
                               "additionally prepare some files for %s, "
                               "i.e. substitute some values)" % pkg_name)
            new_cls.__name__ = orig_cls.__name__

            return new_cls, sub_cmds

        @staticmethod
        def prereq(self):
            """When is reasonable to run this (sub)command"""
            return (
                len(set(pkg_prepare.needs_any_opt).intersection(
                    self.distribution.get_option_dict(pkg_prepare.__name__)
                )) != 0
                and len(set(pkg_prepare.needs_always_opts).difference(
                    self.distribution.get_option_dict(pkg_prepare.__name__)
                )) == 0
            )

        def __init__(self, dist):
            self.pkg_options = tuple(
                key.split('=', 1)[0].replace('-', '_')
                for (key, _, __) in self.user_options
            )
            Command.__init__(self, dist)

        def initialize_options(self):
            # Parameters we want to obtain via setup.cfg, command-line
            # arguments etc. must be prepared as attributes of this object
            if DEBUG: print(DBGPFX + "\tinitialize_options")
            for key in (self.pkg_options + self.needs_any_opt
                        + self.needs_always_opts):
                setattr(self, key, None)
            setattr(self, self.DEV_SWITCH, 0)
            setattr(self, self.BUILD_DEV_SWITCH, 0)

        def finalize_options(self):
            # Obtained parameters are all moved to ``self.pkg_params'' dict
            if DEBUG: print(DBGPFX + "\tfinalize_options")
            for key in self.pkg_options:
                value = getattr(self, key)
                if value is None:
                    raise DistutilsSetupError \
                          ("Missing `%s' value (specify it on command-line"
                           " or in setup.cfg)" % key)
                else:
                    self.pkg_params[key] = value
            # Evaluate all functions within pkg_params (semi-lazy evaluation)
            for key in self.pkg_params:
                value = self.pkg_params[key]
                if hasattr(value, "__call__") or isinstance(value, Callable):
                    self.pkg_params[key] = value(self.pkg_params)
            dist = self.distribution
            dist.data_files = dist.data_files or []
            dist.package_data = dist.package_data or {}
            self.buildonly_files = self.buildonly_files or ()
            self.built_files = self.built_files or ()

        def get_outputs(self):
            return ()  # called from install command for each subcommand

        def run(self):
            if DEBUG: print(DBGPFX + "\trun")
            if getattr(self, self.DEV_SWITCH, 0) \
            or getattr(self, self.BUILD_DEV_SWITCH, 0):
                # Mimic ``develop'' command over "prepared" files
                if DEBUG: print(DBGPFX + "\trun: mimic develop")
                self._pkg_prepare_build()
                self.run_command('build_binary')
                self.run_command('build_ext')
                if getattr(self, self.BUILD_DEV_SWITCH, 0):
                    return
                self._pkg_prepare_install()
                self.run_command('install_data')
                self.run_command('setuptools_develop')
                return
            if self.distribution.get_command_obj('build', create=False):
                # As a part of ``build'' command
                if DEBUG: print(DBGPFX + "\trun: build")
                self._pkg_prepare_build()
            if self.distribution.get_command_obj('install', create=False):
                # As a part of ``install'' command
                if DEBUG: print(DBGPFX + "\trun: install")
                self._pkg_prepare_install()
            else:
                from distutils import log
                log.debug("``pkg_name'' command used with no effect")

        def _pkg_prepare_build(self):
            package_data = self.package_data or {}
            for pkg_name in package_data:
                dst_top = self.distribution.package_dir.get('', '')
                dst_pkg = path_join(
                              dst_top,
                              self.distribution.package_dir.get(pkg_name, pkg_name)
                )
                if DEBUG: print(DBGPFX + "\tbuild dst_pkg %s" % dst_pkg)
                for filedef in package_data[pkg_name]:
                    self._pkg_prepare_file(
                        self.pkg_params[filedef['src']],
                        path_join(dst_pkg, self.pkg_params[filedef['dst']]),
                        filedef.get('substitute', False)
                    )
                    self.distribution.package_data[pkg_name].append(
                        self.pkg_params[filedef['dst']]
                    )
            for filedef in (self.data_files + self.buildonly_files):
                src = self.pkg_params[filedef['src']]
                src_basename = path_basename(src)
                dst_basename = path_basename(self.pkg_params[filedef['dst']])
                substitute = filedef.get('substitute', False)
                if all(c not in src_basename for c in '?*'):
                    if src_basename != dst_basename or substitute:
                        self._pkg_prepare_file(
                            self.pkg_params[filedef['src']],
                            path_join(
                                path_dirname(self.pkg_params[filedef['src']]),
                                dst_basename
                            ),
                            substitute
                        )
                    # eliminate sources from which we prepared files so they
                    # will not end up at build dir and, in turn, installed;
                    # consider only one-level of package at maximum
                    hd, tl = (lambda s, r='': (s, r))(*src.split(sep, 1))
                    if not tl:
                        hd, tl = '', filedef['src']
                    try:
                        self.distribution.package_data.get(hd, []).remove(tl)
                    except ValueError:
                        pass
                else:
                    assert not substitute

        def _pkg_prepare_install(self):
            build_lib = self.get_finalized_command('build_binary').build_lib
            # two-staged copy in case of built_files
            for filedef in self.built_files:
                src = src_orig = self.pkg_params[filedef['src']]
                src = path_basename(src_orig)
                assert src == src_orig, "built_files contains dirs"
                src = path_join(build_lib, src)
                dst = self.pkg_params[filedef['dst']]
                dst = path_join(build_lib, path_basename(dst))
                if src != dst:
                    self._pkg_prepare_file(src, dst)
                self.pkg_params[filedef['src']] = dst
            icmd = self.distribution.get_command_obj('install', create=False)
            for filedef in (self.data_files + self.built_files):
                src = self.pkg_params[filedef['src']]
                dst = self.pkg_params[filedef['dst']]
                no_glob = all(c not in path_basename(src) for c in '?*')
                if dst.startswith(sys_prefix):
                    dst = path_join(icmd.install_base, dst[len(sys_prefix)+1:])
                self.distribution.data_files.append((
                    path_dirname(dst), [
                        path_join(
                            path_dirname(src),
                            path_basename(dst)
                        ),
                    ]
                ) if no_glob else (
                    dst,
                    glob(src)
                ))
                if DEBUG:
                    print(DBGPFX + "\tinstall data_files: %s"
                          % self.distribution.data_files)

        def _pkg_prepare_file(self, src, dst, substitution=False):
            if path_isabs(dst):
                # This function should not attempt doing dangerous things
                # and absolute file destination path is not expected anyway
                # (copying to destination paths is handled by ``install'')
                raise DistutilsSetupError \
                      ("Unexpected attempt to copy file %s to absolute"
                       " location %s" % (src, dst))
            if substitution:
                if DEBUG:
                    print(DBGPFX + "\tSubstituting strings while copying %s -> %s"
                           % (src, dst))
                if self.dry_run:
                    return
                try:
                    with open(src, "r") as fr:
                        with open(dst, "w") as fw:
                            content = fr.read()
                            for key in filter(
                                lambda k: not k.startswith("_")
                                          and k not in self.boolean_options,
                                self.pkg_params
                            ):
                                if DEBUG:
                                    print(DBGPFX + "\tReplace %s -> %s"
                                          % ('@' + key.upper() + '@', self.pkg_params[key]))
                                content = content.replace(
                                              '@' + key.upper() + '@',
                                              self.pkg_params[key]
                                )
                            fw.write(content)
                    copystat(src, dst)
                except IOError as e:
                    raise DistutilsSetupError(str(e))
            else:
                if DEBUG:
                    print(DBGPFX + "\tSimply copying %s -> %s" % (src, dst))
                if self.dry_run:
                    return
                try:
                    copy(src, dst)
                    copystat(src, dst)
                except IOError as e:
                    raise DistutilsSetupError(str(e))
    # class pkg_prepare

    return pkg_prepare


# =========================================================================

self_discovery_plan = ['__project__']
while True:
    pkg, backup_mod, project = {}, None, self_discovery_plan.pop()
    try:
        # XXX check relative import
        backup_mod = sys_modules.get(project)
        if backup_mod:
            if not hasattr(backup_mod, '__path__'):  # not the case for builtins
                continue
            backup_mod = sys_modules.pop(project)
        backup_path, sys_path[:] = sys_path[:], [here]
        pkg = __import__(project, globals=pkg)
        break
    except ImportError:
        sys_path[:] = backup_path
        if backup_mod:
            sys_modules[project] = backup_mod
        if project == '__project__':
            from glob import iglob
            self_discovery_plan.extend(p[:-len('.egg-info')].split('-', 1)[0]
                                       for p in iglob('*.egg-info'))
        if not self_discovery_plan:
            print("Cannot find myself, please help me with __project__ symlink")
            raise
sys_path[:] = backup_path

#pkg_root = path_real('__project__')
pkg_name = pkg.package_name()

# Setup of ``pkg_prepare'' subcommand provided with the items to form options
# for this command;
# each such option requires to be specified (either via command-line or via
# setup.cfg) for correct invocation of this subcommand and after these
# specifications are proceeded, resulting (key, defined value) pairs are added
# into ``pkg_params'' passed as an option to this command in ``setup()''
# so they can be used for string substitutions as well
pkg_prepare = setup_pkg_prepare(pkg_name, (
    ('ccs_flatten',      "location of bundled helper ccs_flatten"),
    ('editor',           "which editor to use if EDITOR env variable not set"),
    ('extplugins_shared',"where plugins are shared between deployments"),
    ('shell_posix',      "POSIX compliant shell for shebangs"),
    ('shell_bashlike',   "bash-like shell (process subst.) for  shebangs"),
    ('hashalgo',         "which hash algorithm to use to generate output name"),
    ('ra_metadata_dir',  "location of RGManager agents/metadata"),
    ('ra_metadata_ext',  "extension used for RGManager agents' metadata"),
    ('report_bugs',      "where to report bugs"),
))

# Contains important values that are then referred to from ``package_data'',
# ``data_files'' etc. definitions within options for ``pkg_prepare'' subcommand
# in ``setup()'', and also serve for strings substituations/interpolations
# (applied to files defined within options for ``pkg_prepare'' subcommand where
# required and except for those starting with underscore)
#
# The right side can be either string or lazily evaluated function returning
# string with this dictionary (dynamically updated with the concrete values
# for parameters passed into ``setup_pkg_prepare'') as a parameter
pkg_param_defaults = 'defaults.py.in'
pkg_params = {
    '__ccs_flatten'        : 'ccs_flatten',
    '__ra_metadata'        : lambda pkg_params:
                             path_norm('ccs-flatten/*.'
                                       + pkg_params['ra_metadata_ext']),
    '__ccs_flatten_config' : path_norm("ccs-flatten/config.h.in"),
    'ccs_flatten_config'   : path_norm("ccs-flatten/config.h"),
    '__defaults'           : path_join(pkg_name, pkg_param_defaults),
    'defaults'             : path_join(pkg_name, 'defaults.py'),
}


# =========================================================================


def cond_require(package, *packages, **preferred):
    packages = (lambda *args: args)(package, *packages)
    for package in packages:
        for preferred_package in preferred:
            sym = preferred[preferred_package]
            try:
                preferred_module = __import__(preferred_package)
                for symbol in ((sym, ) if isinstance(sym, basestring) else sym):
                    if not hasattr(preferred_module, symbol):
                        raise ImportError
            except ImportError:
                return (package, )
    return ()

url_dict = dict(name=pkg_name, ver=pkg.version)
download_url = ''
if PREFER_FORGE == 'github':
    url = 'https://github.com/jnpkrn/{name}'
    if 'git.' in pkg.version:
        download_url = url + '/tarball/' + pkg.version.partition('git.')[-1]
    elif pkg.version.split('+')[-1] != 'a':
        download_url = url + '/tarball/v{ver}'
elif PREFER_FORGE == 'pagure':
    url = 'https://pagure.io/{name}'
    if pkg.version.split('+')[-1] != 'a':
        download_url = 'https://pagure.io/releases/{name}/{name}-{ver}.tar.gz'
else:
    url = 'http://people.redhat.com/jpokorny/pkgs/{name}'
    if pkg.version.split('+')[-1] != 'a':
        download_url = url + '/{name}-{ver}.tar.gz'

setup(

    # STANDARD DISTUTILS ARGUMENTS

    name=pkg_name,
    version=pkg.version,
    description=pkg.description_text(justhead=True),
    url=url.format(**url_dict),
    license=pkg.license,
    author=pkg.author_text(justname=True),
    author_email=pkg.author_text(justname=False),
    #maintainer=pkg.author.partition(", ")[0],
    #maintainer_email=pkg.email.partition(", ")[0],
    download_url=download_url.format(**url_dict),
    #long_description=pkg.description_text(justhead=False),
    long_description=LazyRead('README', lambda c: c.rstrip('\n')),
    classifiers=(
        'Development Status :: 4 - Beta',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: System :: Clustering',
        'Topic :: System :: Systems Administration',
    ),
    keywords='cluster filter framework XML XSLT',

    # ``find_packages'' provided by setuptools avoids the unconvenient
    # listing each subpackage in package hierarchy
    packages=find_packages(
        exclude=('ez_setup', '__project__', '__project__.*', '*.tests'),
    ) + ['.'.join(lib.split(sep)) for lib in
         filter(path_isdir,
                glob(path_join(pkg_name, 'ext-plugins', 'lib-*', '*')))
    ],
    # Following content is also duplicated (in a simplier/more declarative way)
    # in MANIFEST.in which serves for ``setup.py sdist'' command and is
    # necessary due to http://bugs.python.org/issue2279 fixed as of Python 2.7;
    # TODO: MANIFEST.in will be presumably dropped at some point in the future
    # Note: See ``package_data'' under options['pkg_prepare']
    package_data={
        pkg_name: [
            pkg_param_defaults,
            'formats/*/*.rng',
            'formats/*/*.minimal',
            'ext-plugins/PURPOSE',
        ],
    },
    # Note: See ``data_files'' in options for ``pkg_prepare'' subcommand
    #data_files=(),
    # Override ``build'' command handler with local specialized one
    cmdclass=pkg_prepare.inject_cmdclass(
        develop,
        build=(build, build_binary),
        install=(install, install_data),
        setuptools_develop=setuptools_develop,
    ),
    options={
        # Options "passed" to ``pkg_prepare'' subcommand
        pkg_prepare.__name__: dict(
            pkg_params=pkg_params,
            built_files=(
                dict(src='__ccs_flatten', dst='ccs_flatten'),
            ),
            # In addition to standard ``data_files''
            data_files=(
                dict(src='__ra_metadata', dst='ra_metadata_dir'),
            ),
            # Similar to ``data files'' but for files used e.g. for building
            # C extensions
            buildonly_files=(
                dict(src='__ccs_flatten_config', dst='ccs_flatten_config', substitute=True),
                dict(src='__defaults',           dst='defaults',           substitute=True),
            ),

        ),
        build_binary.__name__: dict(
            binaries=(
                Binary(pkg_params['__ccs_flatten'],
                    sources=[path_join('ccs-flatten', f) for f in
                             ('flatten.c',
                              'reslist.c',
                              'resrules.c',
                              'restree.c',
                              'xmlconf.c')],
                    **pkgconfig('libxml-2.0')
                ),
            ),
        ),
    },


    # ADDITIONAL SETUPTOOLS ARGUMENTS

    setup_requires=(
    ),
    install_requires=(
        'lxml',
    ) + cond_require('ordereddict', collections='OrderedDict'),
    # Use pure ``package_data'' value (i.e. do no use MANIFEST.ini or CVS tree);
    # see also comment by ``package data''
    include_package_data=False,

    extras_require={
        'test': cond_require('unittest2', unittest='runner'),
        'test-nose': cond_require('nose', unittest='runner'),
        # see https://bitbucket.org/ned/coveragepy/issues/407#comment-21934227
        'coverage': ('coverage', ) if version_info[:2] != (3, 2)
                                   else ('coverage < 4', ),
    },

    # TODO: uncomment this when ready for tests
    test_suite='{0}.tests'.format(pkg_name),
    tests_require=cond_require('unittest2', unittest='runner'),

    entry_points={
        'console_scripts': (
            '{0} = {0}.__main__'.format(pkg_name),
        ),
    },
)

chdir(prev_cwd)  # restore original CWD (in case we are eval'd or something)



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec, rg2hb_xsl

ccsflat2cibprelude = ('''\
    <!--
        native initscript/systemd unitfile ~ samba
     -->
    <xsl:when test="name() = 'samba'">
        <xsl:choose>
            <xsl:when test="$pcscmd_init_sys = 'systemd'">
''' + (
                ResourceSpec('systemd:smb').xsl_attrs
) + '''\
            </xsl:when>
            <xsl:when test="$pcscmd_init_sys = 'sysvinit'">
''' + (
                ResourceSpec('lsb:smb').xsl_attrs
) + '''\
            </xsl:when>
            <xsl:when test="$pcscmd_init_sys = 'upstart'">
                <!-- XXX could be either lsb, service or upstart
                         (as far as RHEL 6 is concerned),
                         service seems to be most canonical, though -->
''' + (
                ResourceSpec('service:smb').xsl_attrs
) + '''\
            </xsl:when>
            <xsl:otherwise>
''' + (
                ResourceSpec('service:smb').xsl_attrs
) + '''\
                <xsl:comment> %(msg_provider_guess)s </xsl:comment>
                <xsl:message>%(msg_provider_guess)s</xsl:message>
            </xsl:otherwise>
        </xsl:choose>

        <xsl:if test="@config_file or @smbd_options or @nmbd_options">
            <xsl:comment> %(msg_ignored_params)s </xsl:comment>
            <xsl:message>%(msg_ignored_params)s</xsl:message>
        </xsl:if>

        <!-- OPERATIONS -->
        <operations>
''' + (
            rg2hb_xsl('stop', 'shutdown_wait', op=True)
) + '''\
        </operations>
    </xsl:when>
''') % dict(
    msg_provider_guess="WARNING: knowledge of how to start samba on the target"
                       " system missing, assuming `service' provider",
    msg_ignored_params="NOTE: cannot convert samba resource directly, hence"
                       " some configuration bits can be lost"
                       " (config_file,smbd_options, nmbd_options)",
)

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""XSLT helpers"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from .utils_xml import NAMESPACES, XSL, squote
from .utils_2to3 import basestring, iter_items, xrange


NL = '&#xa;'
ALPHA_UPPER = ''.join([chr(i) for i in xrange(ord('A'), ord('Z') + 1)])
ALPHA_LOWER = ALPHA_UPPER.lower()


def xslt_identity(particular_selector=''):
    return '''\
    <xsl:template match="{0}@*|{0}node()"
                  xmlns:xsl="{1}">
        <xsl:copy>
            <xsl:apply-templates select="@*|node()"/>
       </xsl:copy>
    </xsl:template>'''.format(particular_selector, NAMESPACES[XSL])


def xslt_is_member(item, items):
    """Readable item-itemset membership test"""
    items = "\n    '|" + "',\n    '|".join(items) + "',\n    '|'"
    return '''\
    (contains(concat({1}), concat('|', {0}, '|')))'''.format(item, items)


def xslt_boolean(param):
    """Return true/false value as understood within XSL templates"""
    return 'true()' if param else 'false()'


def xslt_params(**d):
    """Convert a provided dictionary into textual XSLT params"""
    ret = ""
    for k, v in iter_items(d):
        ret += '<xsl:param name="{0}" select="{1}"/>\n'.format(
            k, xslt_boolean(v) if isinstance(v, bool)
            else squote(v) if isinstance(v, basestring)
            else v  # int and the like
        )
    return ret


def xslt_id_friendly(inner):
    """Make the passed XPath expression yielding string XML ID friendly"""
    # XXX apostrophe missing
    return '''\
    translate(
        translate(
            {0},
            ' -/:',
            '____'
        ),
        '!&#x22;#$&#x25;&#x26;()*+,;&#x3c;=&#x3e;?@[\]^`{{|}}~',
        ''
    )'''.format(inner)


def xslt_string_mapping(d, what="."):
    """Convert dictionary into procedural mapping application (`xsl:when`s)"""
    ret = []
    for k, v in iter_items(d):
        if not isinstance(v, basestring):
            continue
        ret.append('''\
    <xsl:when test="{0} = '{1}'">'''.format(what, k))
        ret.append('''\
        <xsl:value-of select="'{0}'"/>'''.format(v))
        ret.append('''\
    </xsl:when>''')
    return '\n'.join(ret)


def translate_lower(what="."):
    """Convert textual result of XPath expression to lowercase"""
    return ('''translate({0}, '{1}', '{2}')'''
            .format(what, ALPHA_UPPER, ALPHA_LOWER))



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@config_file',
                     kind='A', desc='configuration file'),
))

###

from ....utils_cib import ResourceSpec, rg2hb_xsl
from ....utils_xml import squote

ccsflat2cibprelude = ('''\
    <!--
        exportfs ~ nfsclient
     -->
    <xsl:when test="name() = 'nfsclient'">
''' + (
        ResourceSpec('ocf:heartbeat:exportfs').xsl_attrs
) + '''

        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
''' + (
            rg2hb_xsl('clientspec', 'target', req=True)
            +
            rg2hb_xsl('directory', 'path', req=True)
            +
            rg2hb_xsl('options')
) + '''\
            <xsl:comment><xsl:value-of select="concat(' ', %(note)s, ' ')"/></xsl:comment>
            <xsl:message><xsl:value-of select="concat(%(note)s)"/></xsl:message>
''' + (
            rg2hb_xsl('unlock_on_stop', 'false', req=abs)
) + '''\
        </instance_attributes>
    </xsl:when>
''') % dict(
    note=', '.join((
        squote("NOTE: explicitly disabling `unlock_on_stop` parameter for"),
        squote(" `exportfs` resource so as to preserve original `nfsclient`"),
        squote(" resource behavior; you may want to enable it, though"),
    ))
)



# -*- coding: UTF-8 -*-
# Copyright 2015 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ....utils_cib import ResourceSpec


ccsflat2cibprelude = '''\
    <!--
        IPAddr2 ~ ip
     -->
    <xsl:when test="name() = 'ip'">
        <xsl:variable name="IpAddress"
                      select="substring-before(@address, '/')"/>
''' + (
        ResourceSpec('ocf:heartbeat:IPaddr2').xsl_attrs
) + '''
        <!-- INSTANCE_ATTRIBUTES -->
        <instance_attributes id="{concat($Prefix, '-ATTRS')}">
            <!-- ip (+ cidr_netmask) ~ address -->
            <nvpair id="{concat($Prefix, '-ATTRS-ip')}"
                    name="ip"
                    value="{@address}">
                <xsl:if test="$IpAddress">
                <xsl:attribute name="value">
                    <xsl:value-of select="$IpAddress"/>
                </xsl:attribute>
                </xsl:if>
            </nvpair>
            <xsl:if test="$IpAddress">
            <nvpair id="{concat($Prefix, '-ATTRS-cidr_netmask')}"
                    name="cidr_netmask"
                    value="{substring-after(@address, '/')}"/>
            </xsl:if>
        </instance_attributes>
    </xsl:when>
'''

###

from ....filters.ccs_artefacts import artefact_cond_ra

ccs_artefacts = ''.join((
    artefact_cond_ra('@address',
                     kind='I', own=True, desc='IP address'),
))



# -*- coding: UTF-8 -*-
# Copyright 2016 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

from ....filters.cib2pcscmd import attrset_xsl

cib2pcscmd = ((
    attrset_xsl("instance_attributes",
                cmd='$pcscmd_pcs, "property set"',
                inform='"set properties for ", @uname, " node"')

) + '''
    <!-- "pcs node utilization" only supported with certain newer
         versions of pcs -->
    <xsl:choose>
        <xsl:when test="$pcscmd_extra_utilization">
''' + (
        attrset_xsl("utilization",
                    cmd='$pcscmd_pcs, "node utilization ", @uname',
                    inform='"set utilization for node: ", @uname')
) + '''
        </xsl:when>
        <xsl:when test="utilization/nvpair">
            <xsl:message>%(utilization_msg)s</xsl:message>
        </xsl:when>
    </xsl:choose>
''') % dict(
    utilization_msg="WARNING: target pcs version does not support utilization"
                    " attributes, hence omitted",
)



# -*- coding: UTF-8 -*-
# Copyright 2018 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

###

needlexml2pcscmd = '''\
    <xsl:value-of select="concat(' ', @ring0_addr)"/>
    <xsl:if test="@ring1_addr">
        <xsl:value-of select="concat(',', @ring1_addr)"/>
    </xsl:if>
    <xsl:if test="@name">
        <!-- https://bugzilla.redhat.com/1183103 -->
        <xsl:message>
            <xsl:value-of select="concat('WARNING: `nodelist.node.name`',
                                         ' value `', ., '` specified,',
                                         ' but current pcs not capable',
                                         ' to set it')"/>
        </xsl:message>
    </xsl:if>
'''



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Format manager"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from .format import formats
from .plugin_registry import PluginManager


class FormatManager(PluginManager):
    """Class responsible for available formats of data to be converted"""
    _default_registry = formats

    @property
    def formats(self):
        return self._plugins



# -*- coding: UTF-8 -*-
# Copyright 2017 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""Pacemaker configuration system/Cluster Information Base (CIB) format"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from os.path import dirname, isabs, join
from sys import modules

from ..format import XML
from ..filters.fmt_cib import fmt_cib_1to2
from ..utils import classproperty


class cib(XML):
    """Pacemaker-based cluster stack configuration (as XML/CIB)

    Also known as Pacemaker Configuration System (cib).
    """
    # XML
    root = 'cib'
    validator_specs = {
        XML.ETREE: 'pacemaker-*.rng'
    }

    _void_file = 'pacemaker-1.2.minimal'

    @classproperty
    def void_file(cls):
        if not isabs(cls._void_file):
            cls._void_file = join(dirname(modules[cls.__module__].__file__),
                                  cls.root, cls._void_file)
        return cls._void_file


class cib_prelude(cib):
    """Private, "unfinished" pacemaker-based cluster stack configuration

    This is a result of ccsflat2cibprelude filter.
    """


class cib_compact(cib):
    """Private, "unfinished" pacemaker-based cluster stack configuration

    This is a result of cibprelude2cibcompact filter.
    """


class cib_final(cib):
    """Public, "finished" pacemaker-based cluster stack configuration

    This is a result of cib2cibfinal filter.
    """


class cib_2(cib):
    """Pacemaker-based cluster stack configuration, 2.x versions

    This may be a result of fmt-cib-1to2 filter going from 1.x versions.
    """
    validator_specs = {
        XML.ETREE: 'pacemaker-2.*.rng'
    }
    compat_contingency = (
        fmt_cib_1to2,  # needs to be real filter reference
    )


class cib_1(cib_2):
    """Pacemaker-based cluster stack configuration, 1.x versions
    """
    validator_specs = {
        XML.ETREE: 'pacemaker-1.*.rng'
    }



# -*- coding: UTF-8 -*-
# Copyright 2014 Red Hat, Inc.
# Part of clufter project
# Licensed under GPLv2+ (a copy included | http://gnu.org/licenses/gpl-2.0.txt)
"""pkgs2installcmd filter"""
__author__ = "Jan Pokorný <jpokorny @at@ Red Hat .dot. com>"

from ..filter import Filter
from ..facts import cmd_pkg_install


@Filter.deco('string-iter', 'command')
def pkgs2installcmd(flt_ctxt, in_obj):
    """Outputs installation command for packages on input"""
    return ('bytestring', cmd_pkg_install(in_obj('native', protect_safe=True),
                          flt_ctxt['system'], flt_ctxt['system_extra']) + '\n')
